from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, TypeVar

import torch
from PIL import Image
from torch import Tensor, device as Device, dtype as DType

from imaginairy.vendored.refiners.foundationals.latent_diffusion.model import LatentDiffusionModel

MAX_STEPS = 1000


@dataclass
class DiffusionTarget:
    size: tuple[int, int]
    offset: tuple[int, int]
    clip_text_embedding: Tensor
    init_latents: Tensor | None = None
    mask_latent: Tensor | None = None
    weight: int = 1
    condition_scale: float = 7.5
    start_step: int = 0
    end_step: int = MAX_STEPS

    def crop(self, tensor: Tensor, /) -> Tensor:
        height, width = self.size
        top_offset, left_offset = self.offset
        return tensor[:, :, top_offset : top_offset + height, left_offset : left_offset + width]

    def paste(self, tensor: Tensor, /, crop: Tensor) -> Tensor:
        height, width = self.size
        top_offset, left_offset = self.offset
        tensor[:, :, top_offset : top_offset + height, left_offset : left_offset + width] = crop
        return tensor


T = TypeVar("T", bound=LatentDiffusionModel)
D = TypeVar("D", bound=DiffusionTarget)


@dataclass
class MultiDiffusion(Generic[T, D], ABC):
    ldm: T

    def __call__(self, x: Tensor, /, noise: Tensor, step: int, targets: list[D]) -> Tensor:
        num_updates = torch.zeros_like(input=x)
        cumulative_values = torch.zeros_like(input=x)

        for target in targets:
            match step:
                case step if step == target.start_step and target.init_latents is not None:
                    noise_view = target.crop(noise)
                    view = self.ldm.scheduler.add_noise(
                        x=target.init_latents,
                        noise=noise_view,
                        step=step,
                    )
                case step if target.start_step <= step <= target.end_step:
                    view = target.crop(x)
                case _:
                    continue
            view = self.diffuse_target(x=view, step=step, target=target)
            weight = target.weight * target.mask_latent if target.mask_latent is not None else target.weight
            num_updates = target.paste(num_updates, crop=target.crop(num_updates) + weight)
            cumulative_values = target.paste(cumulative_values, crop=target.crop(cumulative_values) + weight * view)

        return torch.where(condition=num_updates > 0, input=cumulative_values / num_updates, other=x)

    @abstractmethod
    def diffuse_target(self, x: Tensor, step: int, target: D) -> Tensor:
        ...

    @property
    def steps(self) -> list[int]:
        return self.ldm.steps

    @property
    def device(self) -> Device:
        return self.ldm.device

    @property
    def dtype(self) -> DType:
        return self.ldm.dtype

    def decode_latents(self, x: Tensor) -> Image.Image:
        return self.ldm.lda.decode_latents(x=x)

    @staticmethod
    def generate_offset_grid(size: tuple[int, int], stride: int = 8) -> list[tuple[int, int]]:
        height, width = size

        return [
            (y, x)
            for y in range(0, height, stride)
            for x in range(0, width, stride)
            if y + 64 <= height and x + 64 <= width
        ]



from .dinov2 import (
    DINOv2_base,
    DINOv2_base_reg,
    DINOv2_large,
    DINOv2_large_reg,
    DINOv2_small,
    DINOv2_small_reg,
)
from .vit import (
    ViT,
    ViT_base,
    ViT_large,
    ViT_small,
    ViT_tiny,
)

__all__ = [
    "DINOv2_base",
    "DINOv2_base_reg",
    "DINOv2_large",
    "DINOv2_large_reg",
    "DINOv2_small",
    "DINOv2_small_reg",
    "ViT",
    "ViT_base",
    "ViT_large",
    "ViT_small",
    "ViT_tiny",
]



from collections.abc import Sequence
from enum import Enum, auto

import torch
from jaxtyping import Float, Int
from torch import Tensor, device as Device, dtype as DType, nn

import imaginairy.vendored.refiners.fluxion.layers as fl
from imaginairy.vendored.refiners.fluxion.context import Contexts


class CoordinateEncoder(fl.Chain):
    def __init__(
        self,
        num_positional_features: int = 64,
        scale: float = 1,
        device: Device | str | None = None,
        dtype: DType | None = None,
    ) -> None:
        self.num_positional_features = num_positional_features
        self.scale = scale

        super().__init__(
            fl.Multiply(scale=2, bias=-1),
            fl.Linear(in_features=2, out_features=num_positional_features, bias=False, device=device, dtype=dtype),
            fl.Multiply(scale=2 * torch.pi * self.scale),
            fl.Concatenate(fl.Sin(), fl.Cos(), dim=-1),
        )


class PointType(Enum):
    BACKGROUND = auto()
    FOREGROUND = auto()
    BOX_TOP_LEFT = auto()
    BOX_BOTTOM_RIGHT = auto()
    NOT_A_POINT = auto()


class PointTypeEmbedding(fl.WeightedModule, fl.ContextModule):
    def __init__(self, embedding_dim: int, device: Device | str | None = None, dtype: DType | None = None) -> None:
        super().__init__()
        self.embedding_dim = embedding_dim
        self.weight = nn.Parameter(data=torch.randn(len(PointType), self.embedding_dim, device=device, dtype=dtype))

    def forward(self, type_mask: Int[Tensor, "1 num_points"]) -> Float[Tensor, "1 num_points embedding_dim"]:
        assert isinstance(type_mask, Tensor), "type_mask must be a Tensor."

        embeddings = torch.zeros(*type_mask.shape, self.embedding_dim).to(device=type_mask.device)
        for type_id in PointType:
            mask = type_mask == type_id.value
            embeddings[mask] = self.weight[type_id.value - 1]

        return embeddings


class PointEncoder(fl.Chain):
    def __init__(
        self, embedding_dim: int = 256, scale: float = 1, device: Device | str | None = None, dtype: DType | None = None
    ) -> None:
        assert embedding_dim % 2 == 0, "embedding_dim must be divisible by 2."
        self.embedding_dim = embedding_dim
        self.scale = scale

        super().__init__(
            CoordinateEncoder(num_positional_features=embedding_dim // 2, scale=scale, device=device, dtype=dtype),
            fl.Lambda(func=self.pad),
            fl.Residual(
                fl.UseContext(context="point_encoder", key="type_mask"),
                PointTypeEmbedding(embedding_dim=embedding_dim, device=device, dtype=dtype),
            ),
        )

    def pad(self, x: Tensor) -> Tensor:
        type_mask: Tensor = self.use_context("point_encoder")["type_mask"]
        if torch.any((type_mask == PointType.BOX_TOP_LEFT.value) | (type_mask == PointType.BOX_BOTTOM_RIGHT.value)):
            # Some boxes have been passed: no need to pad in this case
            return x
        type_mask = torch.cat(
            [type_mask, torch.full((type_mask.shape[0], 1), PointType.NOT_A_POINT.value, device=type_mask.device)],
            dim=1,
        )
        self.set_context(context="point_encoder", value={"type_mask": type_mask})
        return torch.cat([x, torch.zeros((x.shape[0], 1, x.shape[-1]), device=x.device)], dim=1)

    def init_context(self) -> Contexts:
        return {
            "point_encoder": {
                "type_mask": None,
            }
        }

    def set_type_mask(self, type_mask: Int[Tensor, "1 num_points"]) -> None:
        self.set_context(context="point_encoder", value={"type_mask": type_mask})

    def get_dense_positional_embedding(
        self, image_embedding_size: tuple[int, int]
    ) -> Float[Tensor, "num_positional_features height width"]:
        coordinate_encoder = self.ensure_find(layer_type=CoordinateEncoder)
        height, width = image_embedding_size
        grid = torch.ones((height, width), device=self.device, dtype=torch.float32)
        y_embedding = grid.cumsum(dim=0) - 0.5
        x_embedding = grid.cumsum(dim=1) - 0.5
        y_embedding = y_embedding / height
        x_embedding = x_embedding / width
        positional_embedding = (
            coordinate_encoder(torch.stack(tensors=[x_embedding, y_embedding], dim=-1))
            .permute(2, 0, 1)
            .unsqueeze(dim=0)
        )
        return positional_embedding

    def points_to_tensor(
        self,
        foreground_points: Sequence[tuple[float, float]] | None = None,
        background_points: Sequence[tuple[float, float]] | None = None,
        not_a_points: Sequence[tuple[float, float]] | None = None,
        box_points: Sequence[Sequence[tuple[float, float]]] | None = None,
    ) -> tuple[Float[Tensor, "1 num_points 2"], Int[Tensor, "1 num_points"]]:
        foreground_points = foreground_points or []
        background_points = background_points or []
        not_a_points = not_a_points or []
        box_points = box_points or []
        top_left_points = [box[0] for box in box_points]
        bottom_right_points = [box[1] for box in box_points]
        coordinates: list[Tensor] = []
        type_ids: list[Tensor] = []

        # Must be in sync with PointType enum
        for type_id, coords_seq in zip(
            PointType, [background_points, foreground_points, top_left_points, bottom_right_points, not_a_points]
        ):
            if len(coords_seq) > 0:
                coords_tensor = torch.tensor(data=list(coords_seq), dtype=torch.float, device=self.device)
                coordinates.append(coords_tensor)
                point_ids = torch.tensor(data=[type_id.value] * len(coords_seq), dtype=torch.int, device=self.device)
                type_ids.append(point_ids)

        all_coordinates = torch.cat(tensors=coordinates, dim=0).unsqueeze(dim=0)
        type_mask = torch.cat(tensors=type_ids, dim=0).unsqueeze(dim=0)

        return all_coordinates, type_mask


class MaskEncoder(fl.Chain):
    def __init__(
        self,
        embedding_dim: int = 256,
        intermediate_channels: int = 16,
        device: Device | str | None = None,
        dtype: DType | None = None,
    ) -> None:
        self.embedding_dim = embedding_dim
        self.intermediate_channels = intermediate_channels
        super().__init__(
            fl.Conv2d(
                in_channels=1,
                out_channels=self.intermediate_channels // 4,
                kernel_size=2,
                stride=2,
                device=device,
                dtype=dtype,
            ),
            fl.LayerNorm2d(channels=self.intermediate_channels // 4, device=device, dtype=dtype),
            fl.GeLU(),
            fl.Conv2d(
                in_channels=self.intermediate_channels // 4,
                out_channels=self.intermediate_channels,
                kernel_size=2,
                stride=2,
                device=device,
                dtype=dtype,
            ),
            fl.LayerNorm2d(channels=self.intermediate_channels, device=device, dtype=dtype),
            fl.GeLU(),
            fl.Conv2d(
                in_channels=self.intermediate_channels,
                out_channels=self.embedding_dim,
                kernel_size=1,
                device=device,
                dtype=dtype,
            ),
        )
        self.register_parameter(
            "no_mask_embedding", nn.Parameter(torch.randn(1, embedding_dim, device=device, dtype=dtype))
        )

    def get_no_mask_dense_embedding(
        self, image_embedding_size: tuple[int, int], batch_size: int = 1
    ) -> Float[Tensor, "batch embedding_dim image_embedding_height image_embedding_width"]:
        return self.no_mask_embedding.reshape(1, -1, 1, 1).expand(
            batch_size, -1, image_embedding_size[0], image_embedding_size[1]
        )



from torch import Tensor

from imaginairy.vendored.refiners.fluxion.layers.module import ContextModule


class Converter(ContextModule):
    """
    A Converter class that adjusts tensor properties based on a parent module's settings.

    This class inherits from `ContextModule` and provides functionality to adjust
    the device and dtype of input tensor(s) to match the parent module's attributes.

    Attributes:
        set_device (bool): If True, matches the device of the input tensor(s) to the parent's device.
        set_dtype (bool): If True, matches the dtype of the input tensor(s) to the parent's dtype.

    Note:
        Ensure the parent module has `device` and `dtype` attributes if `set_device` or `set_dtype` are set to True.
    """

    def __init__(self, set_device: bool = True, set_dtype: bool = True) -> None:
        super().__init__()
        self.set_device = set_device
        self.set_dtype = set_dtype

    def forward(self, *inputs: Tensor) -> tuple[Tensor, ...]:
        parent = self.ensure_parent
        converted_tensors: list[Tensor] = []

        for x in inputs:
            if self.set_device:
                device = parent.device
                assert device is not None, "parent has no device"
                x = x.to(device=device)
            if self.set_dtype:
                dtype = parent.dtype
                assert dtype is not None, "parent has no dtype"
                x = x.to(dtype=dtype)

            converted_tensors.append(x)

        return tuple(converted_tensors)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(set_device={self.set_device}, set_dtype={self.set_dtype})"



"""Utilities for managing AI model weights"""

import os.path

_base_dir = os.path.dirname(os.path.realpath(__file__))

WEIGHT_MAPS_PATH = os.path.join(_base_dir, "weight_maps")
WEIGHT_INFO_PATH = os.path.join(_base_dir, "weight_info")


class MODEL_NAMES:
    SD15 = "stable-diffusion-1-5"


class COMPONENT_NAMES:
    VAE = "vae"
    TEXT_ENCODER = "text"
    UNET = "unet"
    LORA = "lora"


class FORMAT_NAMES:
    COMPVIS = "compvis"
    DIFFUSERS = "diffusers"
    REFINERS = "refiners"


def save_model_info(model_name, component_name, format_name, info_type, data):
    import json

    model_name = model_name.replace("_", "-")
    component_name = component_name.replace("_", "-")
    format_name = format_name.replace("_", "-")
    filename = os.path.join(
        WEIGHT_INFO_PATH,
        f"{model_name}_{component_name}_{format_name}.{info_type}.json",
    )
    with open(filename, "w") as f:
        f.write(json.dumps(data, indent=2))


def prefixes_only(keys):
    new_keys = []
    prev_key = None
    for k in keys:
        new_key = k.rsplit(".", 1)[0]
        if new_key != prev_key:
            new_keys.append(new_key)
        prev_key = new_key
    return new_keys



"""Classes for video sequence transformation"""

import logging
from typing import Optional

import torch
from einops import rearrange, repeat
from torch import nn

from imaginairy.modules.attention import XFORMERS_IS_AVAILABLE
from imaginairy.modules.sgm.attention import (
    CrossAttention,
    FeedForward,
    MemoryEfficientCrossAttention,
    SpatialTransformer,
)
from imaginairy.modules.sgm.diffusionmodules.util import (
    AlphaBlender,
    checkpoint,
    linear,
    timestep_embedding,
)
from imaginairy.utils import exists

logger = logging.getLogger(__name__)


class TimeMixSequential(nn.Sequential):
    def forward(self, x, context=None, timesteps=None):
        for layer in self:
            x = layer(x, context, timesteps)

        return x


class VideoTransformerBlock(nn.Module):
    ATTENTION_MODES = {
        "softmax": CrossAttention,
        "softmax-xformers": MemoryEfficientCrossAttention,
    }

    def __init__(
        self,
        dim,
        n_heads,
        d_head,
        dropout=0.0,
        context_dim=None,
        gated_ff=True,
        checkpoint=False,
        timesteps=None,
        ff_in=False,
        inner_dim=None,
        attn_mode="softmax",
        disable_self_attn=False,
        disable_temporal_crossattention=False,
        switch_temporal_ca_to_sa=False,
    ):
        super().__init__()

        if not XFORMERS_IS_AVAILABLE and attn_mode == "softmax-xformers":
            logger.debug(
                f"Attention mode '{attn_mode}' is not available. Falling back to vanilla attention. "
                f"This is not a problem in Pytorch >= 2.0. FYI, you are running with PyTorch version {torch.__version__}"
            )
            attn_mode = "softmax"

        attn_cls = self.ATTENTION_MODES[attn_mode]

        self.ff_in = ff_in or inner_dim is not None
        if inner_dim is None:
            inner_dim = dim

        assert int(n_heads * d_head) == inner_dim

        self.is_res = inner_dim == dim

        if self.ff_in:
            self.norm_in = nn.LayerNorm(dim)
            self.ff_in = FeedForward(
                dim, dim_out=inner_dim, dropout=dropout, glu=gated_ff
            )

        self.timesteps = timesteps
        self.disable_self_attn = disable_self_attn
        if self.disable_self_attn:
            self.attn1 = attn_cls(
                query_dim=inner_dim,
                heads=n_heads,
                dim_head=d_head,
                context_dim=context_dim,
                dropout=dropout,
            )  # is a cross-attention
        else:
            self.attn1 = attn_cls(
                query_dim=inner_dim, heads=n_heads, dim_head=d_head, dropout=dropout
            )  # is a self-attention

        self.ff = FeedForward(inner_dim, dim_out=dim, dropout=dropout, glu=gated_ff)

        if disable_temporal_crossattention:
            if switch_temporal_ca_to_sa:
                raise ValueError
            else:
                self.attn2 = None
        else:
            self.norm2 = nn.LayerNorm(inner_dim)
            if switch_temporal_ca_to_sa:
                self.attn2 = attn_cls(
                    query_dim=inner_dim, heads=n_heads, dim_head=d_head, dropout=dropout
                )  # is a self-attention
            else:
                self.attn2 = attn_cls(
                    query_dim=inner_dim,
                    context_dim=context_dim,
                    heads=n_heads,
                    dim_head=d_head,
                    dropout=dropout,
                )  # is self-attn if context is none

        self.norm1 = nn.LayerNorm(inner_dim)
        self.norm3 = nn.LayerNorm(inner_dim)
        self.switch_temporal_ca_to_sa = switch_temporal_ca_to_sa

        self.checkpoint = checkpoint
        if self.checkpoint:
            print(f"{self.__class__.__name__} is using checkpointing")

    def forward(
        self,
        x: torch.Tensor,
        context: torch.Tensor = None,
        timesteps: Optional[int] = None,
    ) -> torch.Tensor:
        if self.checkpoint:
            return checkpoint(self._forward, x, context, timesteps)
        else:
            return self._forward(x, context, timesteps=timesteps)

    def _forward(self, x, context=None, timesteps=None):
        assert self.timesteps or timesteps
        assert not (self.timesteps and timesteps) or self.timesteps == timesteps
        timesteps = self.timesteps or timesteps
        B, S, C = x.shape
        x = rearrange(x, "(b t) s c -> (b s) t c", t=timesteps)

        if self.ff_in:
            x_skip = x
            x = self.ff_in(self.norm_in(x))
            if self.is_res:
                x += x_skip

        if self.disable_self_attn:
            x = self.attn1(self.norm1(x), context=context) + x
        else:
            x = self.attn1(self.norm1(x)) + x

        if self.attn2 is not None:
            if self.switch_temporal_ca_to_sa:
                x = self.attn2(self.norm2(x)) + x
            else:
                x = self.attn2(self.norm2(x), context=context) + x
        x_skip = x
        x = self.ff(self.norm3(x))
        if self.is_res:
            x += x_skip

        x = rearrange(
            x, "(b s) t c -> (b t) s c", s=S, b=B // timesteps, c=C, t=timesteps
        )
        return x

    def get_last_layer(self):
        return self.ff.net[-1].weight


class SpatialVideoTransformer(SpatialTransformer):
    def __init__(
        self,
        in_channels,
        n_heads,
        d_head,
        depth=1,
        dropout=0.0,
        use_linear=False,
        context_dim=None,
        use_spatial_context=False,
        timesteps=None,
        merge_strategy: str = "fixed",
        merge_factor: float = 0.5,
        time_context_dim=None,
        ff_in=False,
        checkpoint=False,
        time_depth=1,
        attn_mode="softmax",
        disable_self_attn=False,
        disable_temporal_crossattention=False,
        max_time_embed_period: int = 10000,
    ):
        super().__init__(
            in_channels,
            n_heads,
            d_head,
            depth=depth,
            dropout=dropout,
            attn_type=attn_mode,
            use_checkpoint=checkpoint,
            context_dim=context_dim,
            use_linear=use_linear,
            disable_self_attn=disable_self_attn,
        )
        self.time_depth = time_depth
        self.depth = depth
        self.max_time_embed_period = max_time_embed_period

        time_mix_d_head = d_head
        n_time_mix_heads = n_heads

        time_mix_inner_dim = int(time_mix_d_head * n_time_mix_heads)

        inner_dim = n_heads * d_head
        if use_spatial_context:
            time_context_dim = context_dim

        self.time_stack = nn.ModuleList(
            [
                VideoTransformerBlock(
                    inner_dim,
                    n_time_mix_heads,
                    time_mix_d_head,
                    dropout=dropout,
                    context_dim=time_context_dim,
                    timesteps=timesteps,
                    checkpoint=checkpoint,
                    ff_in=ff_in,
                    inner_dim=time_mix_inner_dim,
                    attn_mode=attn_mode,
                    disable_self_attn=disable_self_attn,
                    disable_temporal_crossattention=disable_temporal_crossattention,
                )
                for _ in range(self.depth)
            ]
        )

        assert len(self.time_stack) == len(self.transformer_blocks)

        self.use_spatial_context = use_spatial_context
        self.in_channels = in_channels

        time_embed_dim = self.in_channels * 4
        self.time_pos_embed = nn.Sequential(
            linear(self.in_channels, time_embed_dim),
            nn.SiLU(),
            linear(time_embed_dim, self.in_channels),
        )

        self.time_mixer = AlphaBlender(
            alpha=merge_factor, merge_strategy=merge_strategy
        )

    def forward(
        self,
        x: torch.Tensor,
        context: Optional[torch.Tensor] = None,
        time_context: Optional[torch.Tensor] = None,
        timesteps: Optional[int] = None,
        image_only_indicator: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        _, _, h, w = x.shape
        x_in = x
        spatial_context = None
        if exists(context):
            spatial_context = context

        if self.use_spatial_context:
            assert (
                context.ndim == 3
            ), f"n dims of spatial context should be 3 but are {context.ndim}"

            time_context = context
            time_context_first_timestep = time_context[::timesteps]
            time_context = repeat(
                time_context_first_timestep, "b ... -> (b n) ...", n=h * w
            )
        elif time_context is not None and not self.use_spatial_context:
            time_context = repeat(time_context, "b ... -> (b n) ...", n=h * w)
            if time_context.ndim == 2:
                time_context = rearrange(time_context, "b c -> b 1 c")

        x = self.norm(x)
        if not self.use_linear:
            x = self.proj_in(x)
        x = rearrange(x, "b c h w -> b (h w) c")
        if self.use_linear:
            x = self.proj_in(x)

        num_frames = torch.arange(timesteps, device=x.device)
        num_frames = repeat(num_frames, "t -> b t", b=x.shape[0] // timesteps)
        num_frames = rearrange(num_frames, "b t -> (b t)")
        t_emb = timestep_embedding(
            num_frames,
            self.in_channels,
            repeat_only=False,
            max_period=self.max_time_embed_period,
        )
        emb = self.time_pos_embed(t_emb)
        emb = emb[:, None, :]

        for it_, (block, mix_block) in enumerate(
            zip(self.transformer_blocks, self.time_stack)
        ):
            x = block(
                x,
                context=spatial_context,
            )

            x_mix = x
            x_mix = x_mix + emb

            x_mix = mix_block(x_mix, context=time_context, timesteps=timesteps)
            x = self.time_mixer(
                x_spatial=x,
                x_temporal=x_mix,
                image_only_indicator=image_only_indicator,
            )
        if self.use_linear:
            x = self.proj_out(x)
        x = rearrange(x, "b (h w) c -> b c h w", h=h, w=w)
        if not self.use_linear:
            x = self.proj_out(x)
        out = x + x_in
        return out



import sys
from collections import defaultdict
from inspect import Parameter, signature
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Any, DefaultDict, Generator, Sequence, TypedDict, TypeVar, cast

from torch import device as Device, dtype as DType
from torch.nn.modules.module import Module as TorchModule

from imaginairy.vendored.refiners.fluxion.context import Context, ContextProvider
from imaginairy.vendored.refiners.fluxion.utils import load_from_safetensors

if TYPE_CHECKING:
    from imaginairy.vendored.refiners.fluxion.layers.chain import Chain

T = TypeVar("T", bound="Module")
TContextModule = TypeVar("TContextModule", bound="ContextModule")
BasicType = str | float | int | bool


class Module(TorchModule):
    _parameters: dict[str, Any]
    _buffers: dict[str, Any]
    _tag: str = ""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, *kwargs)  # type: ignore[reportUnknownMemberType]

    def __getattr__(self, name: str) -> Any:
        return super().__getattr__(name=name)

    def __setattr__(self, name: str, value: Any) -> None:
        return super().__setattr__(name=name, value=value)

    def load_from_safetensors(self, tensors_path: str | Path, strict: bool = True) -> "Module":
        state_dict = load_from_safetensors(tensors_path)
        self.load_state_dict(state_dict, strict=strict)
        return self

    def named_modules(self, *args: Any, **kwargs: Any) -> "Generator[tuple[str, Module], None, None]":  # type: ignore
        return super().named_modules(*args)  # type: ignore

    def to(self: T, device: Device | str | None = None, dtype: DType | None = None) -> T:  # type: ignore
        return super().to(device=device, dtype=dtype)  # type: ignore

    def __str__(self) -> str:
        basic_attributes_str = ", ".join(
            f"{key}={value}" for key, value in self.basic_attributes(init_attrs_only=True).items()
        )
        result = f"{self.__class__.__name__}({basic_attributes_str})"
        return result

    def __repr__(self) -> str:
        tree = ModuleTree(module=self)
        return repr(tree)

    def pretty_print(self, depth: int = -1) -> None:
        tree = ModuleTree(module=self)
        print(tree._generate_tree_repr(tree.root, is_root=True, depth=depth))  # type: ignore[reportPrivateUsage]

    def basic_attributes(self, init_attrs_only: bool = False) -> dict[str, BasicType]:
        """Return a dictionary of basic attributes of the module.

        Basic attributes are public attributes made of basic types (int, float, str, bool) or a sequence of basic types.
        """
        sig = signature(obj=self.__init__)
        init_params = set(sig.parameters.keys()) - {"self"}
        default_values = {k: v.default for k, v in sig.parameters.items() if v.default is not Parameter.empty}

        def is_basic_attribute(key: str, value: Any) -> bool:
            if key.startswith("_"):
                return False

            if isinstance(value, BasicType):
                return True

            if isinstance(value, Sequence) and all(isinstance(y, BasicType) for y in cast(Sequence[Any], value)):
                return True

            return False

        return {
            key: str(object=value)
            for key, value in self.__dict__.items()
            if is_basic_attribute(key=key, value=value)
            and (not init_attrs_only or (key in init_params and value != default_values.get(key)))
        }

    def _show_only_tag(self) -> bool:
        """Whether to show only the tag when printing the module.

        This is useful to distinguish between Chain subclasses that override their forward from one another.
        """
        return False


class ContextModule(Module):
    # we store parent into a one element list to avoid pytorch thinking it's a submodule
    _parent: "list[Chain]"
    _can_refresh_parent: bool = True  # see usage in Adapter and Chain

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, *kwargs)
        self._parent = []

    @property
    def parent(self) -> "Chain | None":
        return self._parent[0] if self._parent else None

    @property
    def ensure_parent(self) -> "Chain":
        assert self._parent, "module is not bound to a Chain"
        return self._parent[0]

    def _set_parent(self, parent: "Chain | None") -> None:
        if not self._can_refresh_parent:
            return
        if parent is None:
            self._parent = []
            return
        # Always insert the module in the Chain first to avoid inconsistencies.
        assert self in iter(parent), f"{self} not in {parent}"
        self._parent = [parent]

    @property
    def provider(self) -> ContextProvider:
        return self.ensure_parent.provider

    def get_parents(self) -> "list[Chain]":
        return self._parent + self._parent[0].get_parents() if self._parent else []

    def use_context(self, context_name: str) -> Context:
        """Retrieve the context object from the module's context provider."""
        context = self.provider.get_context(context_name)
        assert context is not None, f"Context {context_name} not found."
        return context

    def structural_copy(self: TContextModule) -> TContextModule:
        clone = object.__new__(self.__class__)

        not_torch_attributes = [
            key
            for key, value in self.__dict__.items()
            if not key.startswith("_")
            and isinstance(sys.modules.get(type(value).__module__), ModuleType)
            and "torch" not in sys.modules[type(value).__module__].__name__
        ]

        for k in not_torch_attributes:
            setattr(clone, k, getattr(self, k))

        ContextModule.__init__(self=clone)

        return clone


class WeightedModule(Module):
    @property
    def device(self) -> Device:
        return self.weight.device

    @property
    def dtype(self) -> DType:
        return self.weight.dtype

    def __str__(self) -> str:
        return f"{super().__str__().removesuffix(')')}, device={self.device}, dtype={str(self.dtype).removeprefix('torch.')})"


class TreeNode(TypedDict):
    value: str
    class_name: str
    children: list["TreeNode"]


class ModuleTree:
    def __init__(self, module: Module) -> None:
        self.root: TreeNode = self._module_to_tree(module=module)
        self._fold_successive_identical(node=self.root)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(root={self.root['value']})"

    def __repr__(self) -> str:
        return self._generate_tree_repr(self.root, is_root=True, depth=7)

    def __iter__(self) -> Generator[TreeNode, None, None]:
        for child in self.root["children"]:
            yield child

    @classmethod
    def shorten_tree_repr(cls, tree_repr: str, /, line_index: int = 0, max_lines: int = 20) -> str:
        """Shorten the tree representation to a given number of lines around a given line index."""
        lines = tree_repr.split(sep="\n")
        start_idx = max(0, line_index - max_lines // 2)
        end_idx = min(len(lines), line_index + max_lines // 2 + 1)
        return "\n".join(lines[start_idx:end_idx])

    def _generate_tree_repr(
        self, node: TreeNode, /, *, prefix: str = "", is_last: bool = True, is_root: bool = True, depth: int = -1
    ) -> str:
        if depth == 0 and node["children"]:
            return f"{prefix}{'└── ' if is_last else '├── '}{node['value']} ..."

        if depth > 0:
            depth -= 1

        tree_icon: str = "" if is_root else ("└── " if is_last else "├── ")
        counts: DefaultDict[str, int] = defaultdict(int)

        for child in node["children"]:
            counts[child["class_name"]] += 1

        instance_counts: DefaultDict[str, int] = defaultdict(int)
        lines = [f"{prefix}{tree_icon}{node['value']}"]
        new_prefix: str = "    " if is_last else "│   "

        for i, child in enumerate(iterable=node["children"]):
            instance_counts[child["class_name"]] += 1

            if counts[child["class_name"]] > 1:
                child_value = f"{child['value']} #{instance_counts[child['class_name']]}"
            else:
                child_value = child["value"]

            child_str = self._generate_tree_repr(
                {"value": child_value, "class_name": child["class_name"], "children": child["children"]},
                prefix=prefix + new_prefix,
                is_last=i == len(node["children"]) - 1,
                is_root=False,
                depth=depth,
            )

            if child_str:
                lines.append(child_str)

        return "\n".join(lines)

    def _module_to_tree(self, module: Module) -> TreeNode:
        match (module._tag, module._show_only_tag()):  # pyright: ignore[reportPrivateUsage]
            case ("", False):
                value = str(module)
            case (_, True):
                value = f"({module._tag})"  # pyright: ignore[reportPrivateUsage]
            case (_, False):
                value = f"({module._tag}) {module}"  # pyright: ignore[reportPrivateUsage]

        class_name = module.__class__.__name__

        node: TreeNode = {"value": value, "class_name": class_name, "children": []}
        for child in module.children():
            node["children"].append(self._module_to_tree(module=child))  # type: ignore
        return node

    def _fold_successive_identical(self, node: TreeNode) -> None:
        i = 0
        while i < len(node["children"]):
            j = i
            while j < len(node["children"]) and node["children"][i] == node["children"][j]:
                j += 1
            count = j - i
            if count > 1:
                node["children"][i]["value"] += f" (x{count})"
                del node["children"][i + 1 : j]
            self._fold_successive_identical(node=node["children"][i])
            i += 1



"""Functions for creating animations from images."""

import logging
import os.path
from typing import TYPE_CHECKING, List, Sequence

import cv2
import torch

from imaginairy.utils import shrink_list
from imaginairy.utils.img_utils import (
    add_caption_to_image,
    imgpaths_to_imgs,
    model_latents_to_pillow_imgs,
    pillow_img_to_opencv_img,
)

if TYPE_CHECKING:
    from PIL import Image

    from imaginairy.utils.img_utils import LazyLoadingImage


logger = logging.getLogger(__name__)


def make_bounce_animation(
    imgs: "Sequence[Image.Image | LazyLoadingImage | torch.Tensor]",
    outpath: str,
    transition_duration_ms=500,
    start_pause_duration_ms=1000,
    end_pause_duration_ms=2000,
    max_fps=20,
):
    first_img = imgs[0]
    middle_imgs = imgs[1:-1]
    last_img = imgs[-1]

    max_frames = int(round(transition_duration_ms / 1000 * max_fps))
    min_duration = int(1000 / max_fps)
    if middle_imgs:
        progress_duration = int(round(transition_duration_ms / len(middle_imgs)))
    else:
        progress_duration = 0
    progress_duration = max(progress_duration, min_duration)

    middle_imgs = shrink_list(middle_imgs, max_frames)

    frames = [first_img, *middle_imgs, last_img, *list(reversed(middle_imgs))]

    # convert from latents
    converted_frames = _ensure_pillow_images(frames)
    converted_frames = _ensure_images_same_size(converted_frames)

    durations = (
        [start_pause_duration_ms]
        + [progress_duration] * len(middle_imgs)
        + [end_pause_duration_ms]
        + [progress_duration] * len(middle_imgs)
    )
    logger.info(
        f"Making animation with {len(converted_frames)} frames and {progress_duration:.1f}ms per transition frame."
    )
    make_animation(imgs=converted_frames, outpath=outpath, frame_duration_ms=durations)


def _ensure_pillow_images(
    imgs: "List[Image.Image | LazyLoadingImage | torch.Tensor]",
) -> "List[Image.Image]":
    converted_frames: "List[Image.Image]" = []
    for frame in imgs:
        if isinstance(frame, torch.Tensor):
            converted_frames.append(model_latents_to_pillow_imgs(frame)[0])
        else:
            converted_frames.append(frame)  # type: ignore
    return converted_frames


def _ensure_images_same_size(imgs: "List[Image.Image]") -> "List[Image.Image]":
    max_size = max([frame.size for frame in imgs])
    converted_frames = []
    for frame in imgs:
        if frame.size != max_size:
            frame = frame.resize(max_size)
        converted_frames.append(frame)
    return converted_frames


def make_slideshow_animation(
    imgs,
    outpath,
    image_pause_ms=1000,
):
    # convert from latents
    converted_frames = []
    for frame in imgs:
        if isinstance(frame, torch.Tensor):
            frame = model_latents_to_pillow_imgs(frame)[0]
        converted_frames.append(frame)

    durations = [image_pause_ms] * len(converted_frames)

    make_animation(imgs=converted_frames, outpath=outpath, frame_duration_ms=durations)


def make_animation(
    imgs, outpath, frame_duration_ms: int | List[int] = 100, captions=None
):
    imgs = imgpaths_to_imgs(imgs)
    ext = os.path.splitext(outpath)[1].lower().strip(".")

    if captions:
        if len(captions) != len(imgs):
            raise ValueError("Captions and images must be of same length.")
        for img, caption in zip(imgs, captions):
            add_caption_to_image(img, caption)

    if ext == "gif" or ext == "webp":
        make_gif_animation(
            imgs=imgs, outpath=outpath, frame_duration_ms=frame_duration_ms
        )
    elif ext == "mp4":
        make_mp4_animation(
            imgs=imgs, outpath=outpath, frame_duration_ms=frame_duration_ms
        )


def make_gif_animation(imgs, outpath, frame_duration_ms=100, loop=0):
    imgs = imgpaths_to_imgs(imgs)
    imgs[0].save(
        outpath,
        save_all=True,
        append_images=imgs[1:],
        duration=frame_duration_ms,
        loop=loop,
        optimize=False,
    )


def make_mp4_animation(imgs, outpath, frame_duration_ms=50, fps=30, codec="mp4v"):
    imgs = imgpaths_to_imgs(imgs)
    frame_size = imgs[0].size
    fourcc = cv2.VideoWriter_fourcc(*codec)
    out = cv2.VideoWriter(outpath, fourcc, fps, frame_size)
    if not isinstance(frame_duration_ms, list):
        frame_duration_ms = [frame_duration_ms] * len(imgs)
    try:
        for image in select_images_by_duration_at_fps(imgs, frame_duration_ms, fps):
            image = pillow_img_to_opencv_img(image)
            out.write(image)
    finally:
        out.release()


def select_images_by_duration_at_fps(images, durations_ms, fps=30):
    """select the proper image to show for each frame of a video."""
    for i, image in enumerate(images):
        duration = durations_ms[i] / 1000
        num_frames = int(round(duration * fps))
        # print(
        #     f"Showing image {i} for {num_frames} frames for {durations_ms[i]}ms at {fps} fps."
        # )
        for j in range(num_frames):
            yield image



from torch import Tensor

from imaginairy.vendored.refiners.foundationals.clip.image_encoder import CLIPImageEncoderH
from imaginairy.vendored.refiners.foundationals.latent_diffusion.cross_attention import CrossAttentionBlock2d
from imaginairy.vendored.refiners.foundationals.latent_diffusion.image_prompt import ImageProjection, IPAdapter, PerceiverResampler
from imaginairy.vendored.refiners.foundationals.latent_diffusion.stable_diffusion_1.unet import SD1UNet


class SD1IPAdapter(IPAdapter[SD1UNet]):
    def __init__(
        self,
        target: SD1UNet,
        clip_image_encoder: CLIPImageEncoderH | None = None,
        image_proj: ImageProjection | PerceiverResampler | None = None,
        scale: float = 1.0,
        fine_grained: bool = False,
        weights: dict[str, Tensor] | None = None,
    ) -> None:
        clip_image_encoder = clip_image_encoder or CLIPImageEncoderH(device=target.device, dtype=target.dtype)

        if image_proj is None:
            cross_attn_2d = target.ensure_find(CrossAttentionBlock2d)
            image_proj = (
                ImageProjection(
                    clip_image_embedding_dim=clip_image_encoder.output_dim,
                    clip_text_embedding_dim=cross_attn_2d.context_embedding_dim,
                    device=target.device,
                    dtype=target.dtype,
                )
                if not fine_grained
                else PerceiverResampler(
                    latents_dim=cross_attn_2d.context_embedding_dim,
                    num_attention_layers=4,
                    num_attention_heads=12,
                    head_dim=64,
                    num_tokens=16,
                    input_dim=clip_image_encoder.embedding_dim,  # = dim before final projection
                    output_dim=cross_attn_2d.context_embedding_dim,
                    device=target.device,
                    dtype=target.dtype,
                )
            )
        elif fine_grained:
            assert isinstance(image_proj, PerceiverResampler)

        super().__init__(
            target=target,
            clip_image_encoder=clip_image_encoder,
            image_proj=image_proj,
            scale=scale,
            fine_grained=fine_grained,
            weights=weights,
        )



import math
from typing import Any, Callable, Generic, TypeVar

import torch
from torch import Tensor
from torch.fft import fftn, fftshift, ifftn, ifftshift  # type: ignore

import imaginairy.vendored.refiners.fluxion.layers as fl
from imaginairy.vendored.refiners.fluxion.adapters.adapter import Adapter
from imaginairy.vendored.refiners.foundationals.latent_diffusion.stable_diffusion_1.unet import ResidualConcatenator, SD1UNet
from imaginairy.vendored.refiners.foundationals.latent_diffusion.stable_diffusion_xl.unet import SDXLUNet

T = TypeVar("T", bound="SD1UNet | SDXLUNet")
TSDFreeUAdapter = TypeVar("TSDFreeUAdapter", bound="SDFreeUAdapter[Any]")  # Self (see PEP 673)


def fourier_filter(x: Tensor, scale: float = 1, threshold: int = 1) -> Tensor:
    """Fourier filter as introduced in FreeU (https://arxiv.org/abs/2309.11497).

    This version of the method comes from here:
    https://github.com/ChenyangSi/FreeU/blob/main/demo/free_lunch_utils.py#L23
    """
    batch, channels, height, width = x.shape
    dtype = x.dtype
    device = x.device

    if not (math.log2(height).is_integer() and math.log2(width).is_integer()):
        x = x.to(dtype=torch.float32)

    x_freq = fftn(x, dim=(-2, -1))  # type: ignore
    x_freq = fftshift(x_freq, dim=(-2, -1))  # type: ignore
    mask = torch.ones((batch, channels, height, width), device=device)  # type: ignore

    center_row, center_col = height // 2, width // 2  # type: ignore
    mask[..., center_row - threshold : center_row + threshold, center_col - threshold : center_col + threshold] = scale
    x_freq = x_freq * mask  # type: ignore

    x_freq = ifftshift(x_freq, dim=(-2, -1))  # type: ignore
    x_filtered = ifftn(x_freq, dim=(-2, -1)).real  # type: ignore

    return x_filtered.to(dtype=dtype)  # type: ignore


class FreeUBackboneFeatures(fl.Module):
    def __init__(self, backbone_scale: float) -> None:
        super().__init__()
        self.backbone_scale = backbone_scale

    def forward(self, x: Tensor) -> Tensor:
        num_half_channels = x.shape[1] // 2
        x[:, :num_half_channels] = x[:, :num_half_channels] * self.backbone_scale
        return x


class FreeUSkipFeatures(fl.Chain):
    def __init__(self, n: int, skip_scale: float) -> None:
        apply_filter: Callable[[Tensor], Tensor] = lambda x: fourier_filter(x, scale=skip_scale)
        super().__init__(
            fl.UseContext(context="unet", key="residuals").compose(lambda residuals: residuals[n]),
            fl.Lambda(apply_filter),
        )


class FreeUResidualConcatenator(fl.Concatenate):
    def __init__(self, n: int, backbone_scale: float, skip_scale: float) -> None:
        super().__init__(
            FreeUBackboneFeatures(backbone_scale),
            FreeUSkipFeatures(n, skip_scale),
            dim=1,
        )


class SDFreeUAdapter(Generic[T], fl.Chain, Adapter[T]):
    def __init__(self, target: T, backbone_scales: list[float], skip_scales: list[float]) -> None:
        assert len(backbone_scales) == len(skip_scales)
        assert len(backbone_scales) <= len(target.UpBlocks)
        self.backbone_scales = backbone_scales
        self.skip_scales = skip_scales
        with self.setup_adapter(target):
            super().__init__(target)

    def inject(self: TSDFreeUAdapter, parent: fl.Chain | None = None) -> TSDFreeUAdapter:
        for n, (backbone_scale, skip_scale) in enumerate(zip(self.backbone_scales, self.skip_scales)):
            block = self.target.UpBlocks[n]
            concat = block.ensure_find(ResidualConcatenator)
            block.replace(concat, FreeUResidualConcatenator(-n - 2, backbone_scale, skip_scale))
        return super().inject(parent)

    def eject(self) -> None:
        for n in range(len(self.backbone_scales)):
            block = self.target.UpBlocks[n]
            concat = block.ensure_find(FreeUResidualConcatenator)
            block.replace(concat, ResidualConcatenator(-n - 2))
        super().eject()



"""Refinement modules for image generation"""

import logging
import math
from functools import lru_cache
from typing import Any, List, Literal

import numpy as np
import torch
from PIL import Image
from torch import Tensor, device as Device, dtype as DType, nn
from torch.nn import functional as F

import imaginairy.vendored.refiners.fluxion.layers as fl
from imaginairy import config
from imaginairy.schema import WeightedPrompt
from imaginairy.utils.downloads import get_cached_url_path
from imaginairy.utils.feather_tile import rebuild_image, tile_image
from imaginairy.vendored.refiners.fluxion.layers.attentions import (
    ScaledDotProductAttention,
)
from imaginairy.vendored.refiners.fluxion.layers.chain import ChainError
from imaginairy.vendored.refiners.fluxion.utils import image_to_tensor, interpolate
from imaginairy.vendored.refiners.foundationals.clip.text_encoder import (
    CLIPTextEncoderL,
)
from imaginairy.vendored.refiners.foundationals.latent_diffusion import (
    SD1IPAdapter,
    SDXLIPAdapter,
)
from imaginairy.vendored.refiners.foundationals.latent_diffusion.model import (
    TLatentDiffusionModel,
)
from imaginairy.vendored.refiners.foundationals.latent_diffusion.schedulers.ddim import (
    DDIM,
)
from imaginairy.vendored.refiners.foundationals.latent_diffusion.schedulers.scheduler import (
    Scheduler,
)
from imaginairy.vendored.refiners.foundationals.latent_diffusion.self_attention_guidance import (
    SelfAttentionMap,
)
from imaginairy.vendored.refiners.foundationals.latent_diffusion.stable_diffusion_1.controlnet import (
    Controlnet,
    SD1ControlnetAdapter,
)
from imaginairy.vendored.refiners.foundationals.latent_diffusion.stable_diffusion_1.model import (
    SD1Autoencoder,
    SD1UNet,
    StableDiffusion_1 as RefinerStableDiffusion_1,
    StableDiffusion_1_Inpainting as RefinerStableDiffusion_1_Inpainting,
)
from imaginairy.vendored.refiners.foundationals.latent_diffusion.stable_diffusion_xl.model import (
    SDXLAutoencoder,
    StableDiffusion_XL as RefinerStableDiffusion_XL,
)
from imaginairy.vendored.refiners.foundationals.latent_diffusion.stable_diffusion_xl.text_encoder import (
    DoubleTextEncoder,
)
from imaginairy.vendored.refiners.foundationals.latent_diffusion.stable_diffusion_xl.unet import (
    SDXLUNet,
)
from imaginairy.weight_management.conversion import cast_weights
from imaginairy.weight_management.translators import (
    diffusers_ip_adapter_plus_sd15_to_refiners_translator,
    diffusers_ip_adapter_plus_sdxl_to_refiners_translator,
    diffusers_ip_adapter_sd15_to_refiners_translator,
    diffusers_ip_adapter_sdxl_to_refiners_translator,
    transformers_image_encoder_to_refiners_translator,
)

logger = logging.getLogger(__name__)

TileModeType = Literal["", "x", "y", "xy"]


def _tile_mode_conv2d_conv_forward(
    self,
    input: torch.Tensor,  # noqa
    weight: torch.Tensor,
    bias: torch.Tensor,
):
    if self.padding_mode_x == self.padding_mode_y:
        self.padding_mode = self.padding_mode_x
        return self._orig_conv_forward(input, weight, bias)

    w1 = F.pad(input, self.padding_x, mode=self.padding_modeX)
    del input

    w2 = F.pad(w1, self.padding_y, mode=self.padding_modeY)
    del w1

    return F.conv2d(w2, weight, bias, self.stride, (0, 0), self.dilation, self.groups)


class TileModeMixin(nn.Module):
    def set_tile_mode(self, tile_mode: TileModeType = ""):
        """
        For creating seamless tile images.

        Args:
            tile_mode: One of "", "x", "y", "xy". If "x", the image will be tiled horizontally. If "y", the image will be
                tiled vertically. If "xy", the image will be tiled both horizontally and vertically.
        """
        padding_mode_x = "circular" if "x" in tile_mode else "constant"
        padding_mode_y = "circular" if "y" in tile_mode else "constant"
        for m in self.modules():
            if not isinstance(m, nn.Conv2d):
                continue
            if not hasattr(m, "_orig_conv_forward"):
                # patch with a function that can handle tiling in a single direction
                m._initial_padding_mode = m.padding_mode  # type: ignore
                m._orig_conv_forward = m._conv_forward  # type: ignore
                m._conv_forward = _tile_mode_conv2d_conv_forward.__get__(m, nn.Conv2d)  # type: ignore
            m.padding_mode_x = padding_mode_x  # type: ignore
            m.padding_mode_y = padding_mode_y  # type: ignore
            rprt: list[int] = m._reversed_padding_repeated_twice
            m.padding_x = (rprt[0], rprt[1], 0, 0)  # type: ignore
            m.padding_y = (0, 0, rprt[2], rprt[3])  # type: ignore


class SD1ImagePromptMixin(nn.Module):
    def _get_ip_adapter(self, model_type: str):
        valid_model_types = ["normal", "plus", "plus-face"]
        if model_type not in valid_model_types:
            msg = f"IP Adapter model_type must be one of {valid_model_types}"
            raise ValueError(msg)

        ip_adapter_weights_path = get_cached_url_path(
            config.IP_ADAPTER_WEIGHT_LOCATIONS["sd15"][model_type]
        )
        clip_image_weights_path = get_cached_url_path(config.SD21_UNCLIP_WEIGHTS_URL)
        if "plus" in model_type:
            ip_adapter_weight_translator = (
                diffusers_ip_adapter_plus_sd15_to_refiners_translator()
            )
        else:
            ip_adapter_weight_translator = (
                diffusers_ip_adapter_sd15_to_refiners_translator()
            )
        clip_image_weight_translator = (
            transformers_image_encoder_to_refiners_translator()
        )

        ip_adapter = SD1IPAdapter(
            target=self.unet,
            weights=ip_adapter_weight_translator.load_and_translate_weights(
                ip_adapter_weights_path
            ),
            fine_grained="plus" in model_type,
        )
        ip_adapter.clip_image_encoder.load_state_dict(
            clip_image_weight_translator.load_and_translate_weights(
                clip_image_weights_path
            ),
            assign=True,
        )
        ip_adapter.to(device=self.unet.device, dtype=self.unet.dtype)
        ip_adapter.clip_image_encoder.to(device=self.unet.device, dtype=self.unet.dtype)
        return ip_adapter

    def set_image_prompt(
        self, images: list[Image.Image], scale: float, model_type: str = "normal"
    ):
        ip_adapter = self._get_ip_adapter(model_type)
        ip_adapter.inject()

        ip_adapter.set_scale(scale)
        image_embeddings = []
        for image in images:
            image_embedding = ip_adapter.compute_clip_image_embedding(
                ip_adapter.preprocess_image(image).to(device=self.unet.device)
            )
            image_embeddings.append(image_embedding)

        clip_image_embedding = sum(image_embeddings) / len(image_embeddings)

        ip_adapter.set_clip_image_embedding(clip_image_embedding)


class StableDiffusion_1(TileModeMixin, SD1ImagePromptMixin, RefinerStableDiffusion_1):
    def __init__(
        self,
        unet: SD1UNet | None = None,
        lda: SD1Autoencoder | None = None,
        clip_text_encoder: CLIPTextEncoderL | None = None,
        scheduler: Scheduler | None = None,
        device: Device | str | None = "cpu",
        dtype: DType = torch.float32,
    ) -> None:
        unet = unet or SD1UNet(in_channels=4)
        lda = lda or SD1Autoencoder()
        clip_text_encoder = clip_text_encoder or CLIPTextEncoderL()
        scheduler = scheduler or DDIM(num_inference_steps=50)
        fl.Module.__init__(self)

        # all this is to allow us to make structural copies without unnecessary device or dtype shuffeling
        # since default behavior was to put everything on the same device and dtype and we want the option to
        # not alter them from whatever they're already set to
        self.unet = unet
        self.lda = lda
        self.clip_text_encoder = clip_text_encoder
        self.scheduler = scheduler
        to_kwargs: dict[str, Any] = {}

        if device is not None:
            device = device if isinstance(device, Device) else Device(device=device)
            to_kwargs["device"] = device
        if dtype is not None:
            to_kwargs["dtype"] = dtype

        self.device = device  # type: ignore
        self.dtype = dtype

        if to_kwargs:
            self.unet = unet.to(**to_kwargs)
            self.lda = lda.to(**to_kwargs)
            self.clip_text_encoder = clip_text_encoder.to(**to_kwargs)
            self.scheduler = scheduler.to(**to_kwargs)

    def calculate_text_conditioning_kwargs(
        self,
        positive_prompts: List[WeightedPrompt],
        negative_prompts: List[WeightedPrompt],
        positive_conditioning_override: Tensor | None = None,
    ):
        import torch

        from imaginairy.utils.log_utils import log_conditioning

        neutral_conditioning = self.prompts_to_embeddings(negative_prompts)
        log_conditioning(neutral_conditioning, "neutral conditioning")

        if positive_conditioning_override is None:
            positive_conditioning = self.prompts_to_embeddings(positive_prompts)
        else:
            positive_conditioning = positive_conditioning_override
        log_conditioning(positive_conditioning, "positive conditioning")

        clip_text_embedding = torch.cat(
            tensors=(neutral_conditioning, positive_conditioning), dim=0
        )
        return {"clip_text_embedding": clip_text_embedding}

    def prompts_to_embeddings(self, prompts: List[WeightedPrompt]) -> Tensor:
        import torch

        total_weight = sum(wp.weight for wp in prompts)
        if str(self.clip_text_encoder.device) == "cpu":
            self.clip_text_encoder = self.clip_text_encoder.to(dtype=torch.float32)
        conditioning = sum(
            self.clip_text_encoder(wp.text) * (wp.weight / total_weight)
            for wp in prompts
        )

        return conditioning


class SDXLImagePromptMixin(nn.Module):
    def _get_ip_adapter(self, model_type: str):
        valid_model_types = ["normal", "plus", "plus-face"]
        if model_type not in valid_model_types:
            msg = f"IP Adapter model_type must be one of {valid_model_types}"
            raise ValueError(msg)

        ip_adapter_weights_path = get_cached_url_path(
            config.IP_ADAPTER_WEIGHT_LOCATIONS["sdxl"][model_type]
        )
        clip_image_weights_path = get_cached_url_path(config.SD21_UNCLIP_WEIGHTS_URL)
        if "plus" in model_type:
            ip_adapter_weight_translator = (
                diffusers_ip_adapter_plus_sdxl_to_refiners_translator()
            )
        else:
            ip_adapter_weight_translator = (
                diffusers_ip_adapter_sdxl_to_refiners_translator()
            )
        clip_image_weight_translator = (
            transformers_image_encoder_to_refiners_translator()
        )

        ip_adapter = SDXLIPAdapter(
            target=self.unet,
            weights=ip_adapter_weight_translator.load_and_translate_weights(
                ip_adapter_weights_path
            ),
            fine_grained="plus" in model_type,
        )
        ip_adapter.clip_image_encoder.load_state_dict(
            clip_image_weight_translator.load_and_translate_weights(
                clip_image_weights_path
            ),
            assign=True,
        )
        ip_adapter.to(device=self.unet.device, dtype=self.unet.dtype)
        ip_adapter.clip_image_encoder.to(device=self.unet.device, dtype=self.unet.dtype)
        return ip_adapter

    def set_image_prompt(
        self, images: list[Image.Image], scale: float, model_type: str = "normal"
    ):
        ip_adapter = self._get_ip_adapter(model_type)
        ip_adapter.inject()

        ip_adapter.set_scale(scale)
        image_embeddings = []
        for image in images:
            image_embedding = ip_adapter.compute_clip_image_embedding(
                ip_adapter.preprocess_image(image).to(device=self.unet.device)
            )
            image_embeddings.append(image_embedding)

        clip_image_embedding = sum(image_embeddings) / len(image_embeddings)

        ip_adapter.set_clip_image_embedding(clip_image_embedding)


class StableDiffusion_XL(
    TileModeMixin, SDXLImagePromptMixin, RefinerStableDiffusion_XL
):
    def __init__(
        self,
        unet: SDXLUNet | None = None,
        lda: SDXLAutoencoder | None = None,
        clip_text_encoder: DoubleTextEncoder | None = None,
        scheduler: Scheduler | None = None,
        device: Device | str | None = "cpu",
        dtype: DType | None = None,
    ) -> None:
        unet = unet or SDXLUNet(in_channels=4)
        lda = lda or SDXLAutoencoder()
        clip_text_encoder = clip_text_encoder or DoubleTextEncoder()
        scheduler = scheduler or DDIM(num_inference_steps=30)
        fl.Module.__init__(self)

        # all this is to allow us to make structural copies without unnecessary device or dtype shuffeling
        # since default behavior was to put everything on the same device and dtype and we want the option to
        # not alter them from whatever they're already set to
        self.unet = unet
        self.lda = lda
        self.clip_text_encoder = clip_text_encoder
        self.scheduler = scheduler
        to_kwargs: dict[str, Any] = {}

        if device is not None:
            device = device if isinstance(device, Device) else Device(device=device)
            to_kwargs["device"] = device
        if dtype is not None:
            to_kwargs["dtype"] = dtype

        self.device = device  # type: ignore
        self.dtype = dtype  # type: ignore
        self.unet = unet
        self.lda = lda
        self.clip_text_encoder = clip_text_encoder
        self.scheduler = scheduler
        if to_kwargs:
            self.unet = self.unet.to(**to_kwargs)
            self.lda = self.lda.to(**to_kwargs)
            self.clip_text_encoder = self.clip_text_encoder.to(**to_kwargs)
            self.scheduler = self.scheduler.to(**to_kwargs)

    def forward(  # type: ignore
        self,
        x: Tensor,
        step: int,
        *,
        clip_text_embedding: Tensor,
        pooled_text_embedding: Tensor,
        time_ids: Tensor | None = None,
        condition_scale: float = 5.0,
        **kwargs: Tensor,
    ) -> Tensor:
        time_ids = time_ids or self.default_time_ids
        return super().forward(
            x=x,
            step=step,
            clip_text_embedding=clip_text_embedding,
            pooled_text_embedding=pooled_text_embedding,
            time_ids=time_ids,
            condition_scale=condition_scale,
            **kwargs,
        )

    def structural_copy(self: TLatentDiffusionModel) -> TLatentDiffusionModel:
        logger.debug("Making structural copy of StableDiffusion_XL model")

        sd = self.__class__(
            unet=self.unet.structural_copy(),
            lda=self.lda.structural_copy(),
            clip_text_encoder=self.clip_text_encoder,
            scheduler=self.scheduler,
            device=self.device,
            dtype=None,  # type: ignore
        )
        logger.debug(
            f"dtype: {sd.dtype} unet-dtype:{sd.unet.dtype} lda-dtype:{sd.lda.dtype} text-encoder-dtype:{sd.clip_text_encoder.dtype} scheduler-dtype:{sd.scheduler.dtype}"
        )
        return sd

    def calculate_text_conditioning_kwargs(
        self,
        positive_prompts: List[WeightedPrompt],
        negative_prompts: List[WeightedPrompt],
        positive_conditioning_override: Tensor | None = None,
    ):
        import torch

        from imaginairy.utils.log_utils import log_conditioning

        (
            neutral_clip_text_embedding,
            neutral_pooled_text_embedding,
        ) = self.prompts_to_embeddings(negative_prompts)
        log_conditioning(neutral_clip_text_embedding, "neutral_clip_text_embedding")
        log_conditioning(neutral_pooled_text_embedding, "neutral_pooled_text_embedding")

        (
            positive_clip_text_embedding,
            positive_pooled_text_embedding,
        ) = self.prompts_to_embeddings(positive_prompts)
        log_conditioning(positive_clip_text_embedding, "positive_clip_text_embedding")
        log_conditioning(
            positive_pooled_text_embedding, "positive_pooled_text_embedding"
        )

        return {
            "clip_text_embedding": torch.cat(
                tensors=(neutral_clip_text_embedding, positive_clip_text_embedding),
                dim=0,
            ),
            "pooled_text_embedding": torch.cat(
                tensors=(neutral_pooled_text_embedding, positive_pooled_text_embedding),
                dim=0,
            ),
        }

    def prompts_to_embeddings(
        self, prompts: List[WeightedPrompt]
    ) -> tuple[Tensor, Tensor]:
        import torch

        total_weight = sum(wp.weight for wp in prompts)
        if str(self.clip_text_encoder.device) == "cpu":
            self.clip_text_encoder = self.clip_text_encoder.to(dtype=torch.float32)

        embeddings = [self.clip_text_encoder(wp.text) for wp in prompts]
        clip_text_embedding = (
            sum(emb[0] * wp.weight for emb, wp in zip(embeddings, prompts))
            / total_weight
        )
        pooled_text_embedding = (
            sum(emb[1] * wp.weight for emb, wp in zip(embeddings, prompts))
            / total_weight
        )

        return clip_text_embedding, pooled_text_embedding  # type: ignore


class StableDiffusion_1_Inpainting(
    TileModeMixin, SD1ImagePromptMixin, RefinerStableDiffusion_1_Inpainting
):
    def compute_self_attention_guidance(
        self,
        x: Tensor,
        noise: Tensor,
        step: int,
        *,
        clip_text_embedding: Tensor,
        **kwargs: Tensor,
    ) -> Tensor:
        sag = self._find_sag_adapter()
        assert sag is not None
        assert self.mask_latents is not None
        assert self.target_image_latents is not None

        degraded_latents = sag.compute_degraded_latents(
            scheduler=self.scheduler,
            latents=x,
            noise=noise,
            step=step,
            classifier_free_guidance=True,
        )

        negative_embedding, _ = clip_text_embedding.chunk(2)
        timestep = self.scheduler.timesteps[step].unsqueeze(dim=0)
        self.set_unet_context(
            timestep=timestep, clip_text_embedding=negative_embedding, **kwargs
        )
        x = torch.cat(
            tensors=(degraded_latents, self.mask_latents, self.target_image_latents),
            dim=1,
        )
        if "ip_adapter" in self.unet.provider.contexts:
            # this implementation is a bit hacky, it should be refactored in the future
            ip_adapter_context = self.unet.use_context("ip_adapter")
            image_embedding_copy = ip_adapter_context["clip_image_embedding"].clone()
            ip_adapter_context["clip_image_embedding"], _ = ip_adapter_context[
                "clip_image_embedding"
            ].chunk(2)
            degraded_noise = self.unet(x)
            ip_adapter_context["clip_image_embedding"] = image_embedding_copy
        else:
            degraded_noise = self.unet(x)

        return sag.scale * (noise - degraded_noise)

    def calculate_text_conditioning_kwargs(
        self,
        positive_prompts: List[WeightedPrompt],
        negative_prompts: List[WeightedPrompt],
        positive_conditioning_override: Tensor | None = None,
    ):
        import torch

        from imaginairy.utils.log_utils import log_conditioning

        neutral_conditioning = self.prompts_to_embeddings(negative_prompts)
        log_conditioning(neutral_conditioning, "neutral conditioning")

        if positive_conditioning_override is None:
            positive_conditioning = self.prompts_to_embeddings(positive_prompts)
        else:
            positive_conditioning = positive_conditioning_override
        log_conditioning(positive_conditioning, "positive conditioning")

        clip_text_embedding = torch.cat(
            tensors=(neutral_conditioning, positive_conditioning), dim=0
        )
        return {"clip_text_embedding": clip_text_embedding}

    def prompts_to_embeddings(self, prompts: List[WeightedPrompt]) -> Tensor:
        import torch

        total_weight = sum(wp.weight for wp in prompts)
        if str(self.clip_text_encoder.device) == "cpu":  # type: ignore
            self.clip_text_encoder = self.clip_text_encoder.to(dtype=torch.float32)  # type: ignore
        conditioning = sum(
            self.clip_text_encoder(wp.text) * (wp.weight / total_weight)
            for wp in prompts
        )

        return conditioning


class StableDiffusion_XL_Inpainting(StableDiffusion_XL):
    def __init__(
        self,
        unet: SDXLUNet | None = None,
        lda: SDXLAutoencoder | None = None,
        clip_text_encoder: DoubleTextEncoder | None = None,
        scheduler: Scheduler | None = None,
        device: Device | str | None = "cpu",
        dtype: DType | None = None,
    ) -> None:
        self.mask_latents: Tensor | None = None
        self.target_image_latents: Tensor | None = None
        super().__init__(
            unet=unet,
            lda=lda,
            clip_text_encoder=clip_text_encoder,
            scheduler=scheduler,
            device=device,
            dtype=dtype,
        )

    def forward(
        self,
        x: Tensor,
        step: int,
        *,
        clip_text_embedding: Tensor,
        pooled_text_embedding: Tensor,
        time_ids: Tensor | None = None,
        condition_scale: float = 5.0,
        **_: Tensor,
    ) -> Tensor:
        assert self.mask_latents is not None
        assert self.target_image_latents is not None
        x = torch.cat(tensors=(x, self.mask_latents, self.target_image_latents), dim=1)
        return super().forward(
            x=x,
            step=step,
            clip_text_embedding=clip_text_embedding,
            pooled_text_embedding=pooled_text_embedding,
            time_ids=time_ids,
            condition_scale=condition_scale,
        )

    def set_inpainting_conditions(
        self,
        target_image: Image.Image,
        mask: Image.Image,
        latents_size: tuple[int, int] = (64, 64),
    ) -> tuple[Tensor, Tensor]:
        target_image = target_image.convert(mode="RGB")
        mask = mask.convert(mode="L")

        mask_tensor = torch.tensor(
            data=np.array(object=mask).astype(dtype=np.float32) / 255.0
        ).to(device=self.device)
        mask_tensor = (
            (mask_tensor > 0.5)
            .unsqueeze(dim=0)
            .unsqueeze(dim=0)
            .to(dtype=self.unet.dtype)
        )

        self.mask_latents = interpolate(x=mask_tensor, factor=torch.Size(latents_size))

        init_image_tensor = (
            image_to_tensor(
                image=target_image, device=self.device, dtype=self.unet.dtype
            )
            * 2
            - 1
        )
        masked_init_image = init_image_tensor * (1 - mask_tensor)
        self.target_image_latents = self.lda.encode(
            x=masked_init_image.to(dtype=self.lda.dtype)
        )
        assert self.target_image_latents is not None
        self.target_image_latents = self.target_image_latents.to(dtype=self.unet.dtype)

        return self.mask_latents, self.target_image_latents  # type: ignore

    def compute_self_attention_guidance(
        self,
        x: Tensor,
        noise: Tensor,
        step: int,
        *,
        clip_text_embedding: Tensor,
        pooled_text_embedding: Tensor,
        time_ids: Tensor,
        **kwargs: Tensor,
    ) -> Tensor:
        sag = self._find_sag_adapter()
        assert sag is not None
        assert self.mask_latents is not None
        assert self.target_image_latents is not None

        degraded_latents = sag.compute_degraded_latents(
            scheduler=self.scheduler,
            latents=x,
            noise=noise,
            step=step,
            classifier_free_guidance=True,
        )

        negative_embedding, _ = clip_text_embedding.chunk(2)
        negative_pooled_embedding, _ = pooled_text_embedding.chunk(2)
        timestep = self.scheduler.timesteps[step].unsqueeze(dim=0)
        time_ids, _ = time_ids.chunk(2)
        self.set_unet_context(
            timestep=timestep,
            clip_text_embedding=negative_embedding,
            pooled_text_embedding=negative_pooled_embedding,
            time_ids=time_ids,
            **kwargs,
        )
        x = torch.cat(
            tensors=(degraded_latents, self.mask_latents, self.target_image_latents),
            dim=1,
        )
        if "ip_adapter" in self.unet.provider.contexts:
            # this implementation is a bit hacky, it should be refactored in the future
            ip_adapter_context = self.unet.use_context("ip_adapter")
            image_embedding_copy = ip_adapter_context["clip_image_embedding"].clone()
            ip_adapter_context["clip_image_embedding"], _ = ip_adapter_context[
                "clip_image_embedding"
            ].chunk(2)
            degraded_noise = self.unet(x)
            ip_adapter_context["clip_image_embedding"] = image_embedding_copy
        else:
            degraded_noise = self.unet(x)

        return sag.scale * (noise - degraded_noise)


class SlicedEncoderMixin(nn.Module):
    max_chunk_size = 2048
    min_chunk_size = 32

    def encode(self, x: Tensor) -> Tensor:
        return self.sliced_encode(x)

    def sliced_encode(self, x: Tensor, chunk_size: int = 128 * 8) -> Tensor:
        """
        Encodes the image in slices (for lower memory usage).
        """
        b, c, h, w = x.size()
        final_tensor = torch.zeros(
            [1, 4, math.floor(h / 8), math.floor(w / 8)], device=x.device
        )
        overlap_pct = 0.5
        encoder = self[0]  # type: ignore
        for x_img in x.split(1):
            chunks = tile_image(
                x_img, tile_size=chunk_size, overlap_percent=overlap_pct
            )
            encoded_chunks = [encoder(ic) * self.encoder_scale for ic in chunks]

            final_tensor = rebuild_image(
                encoded_chunks,
                base_img=final_tensor,
                tile_size=chunk_size // 8,
                overlap_percent=overlap_pct,
            )

        return final_tensor

    def decode(self, x):
        while self.__class__.max_chunk_size > self.__class__.min_chunk_size:
            if self.max_chunk_size**2 > x.shape[2] * x.shape[3]:
                try:
                    return self.decode_all_at_once(x)
                except ChainError as e:
                    if "OutOfMemoryError" not in str(e):
                        raise
                    new_size = int(math.sqrt(x.shape[2] * x.shape[3])) // 2
                    # make sure it's an even number
                    new_size = new_size - (new_size % 2)
                    self.__class__.max_chunk_size = new_size
                    logger.info(
                        f"Ran out of memory. Trying tiled decode with chunk size {self.__class__.max_chunk_size}"
                    )
            else:
                try:
                    return self.decode_sliced(x, chunk_size=self.max_chunk_size)
                except ChainError as e:
                    if "OutOfMemoryError" not in str(e):
                        raise
                    self.__class__.max_chunk_size = self.max_chunk_size // 2
                    # make sure it's an even number
                    self.__class__.max_chunk_size -= self.__class__.max_chunk_size % 2
                    self.__class__.max_chunk_size = max(
                        self.__class__.max_chunk_size, self.__class__.min_chunk_size
                    )
                    logger.info(
                        f"Ran out of memory. Trying tiled decode with chunk size {self.__class__.max_chunk_size}"
                    )
        raise RuntimeError("Could not decode image")

    def decode_all_at_once(self, x: Tensor) -> Tensor:
        decoder = self[1]  # type: ignore
        x = decoder(x / self.encoder_scale)
        return x

    def decode_sliced(self, x, chunk_size=128):
        """
        decodes the tensor in slices.

        This results in image portions that don't exactly match, so we overlap, feather, and merge to reduce
        (but not completely eliminate) impact.
        """
        b, c, h, w = x.size()
        final_tensor = torch.zeros([1, 3, h * 8, w * 8], device=x.device)
        for x_latent in x.split(1):
            decoded_chunks = []
            overlap_pct = 0.5
            chunks = tile_image(
                x_latent, tile_size=chunk_size, overlap_percent=overlap_pct
            )

            for latent_chunk in chunks:
                # latent_chunk = self.post_quant_conv(latent_chunk)
                dec = self.decode_all_at_once(latent_chunk)
                decoded_chunks.append(dec)
            final_tensor = rebuild_image(
                decoded_chunks,
                base_img=final_tensor,
                tile_size=chunk_size * 8,
                overlap_percent=overlap_pct,
            )

            return final_tensor


class SD1AutoencoderSliced(SlicedEncoderMixin, SD1Autoencoder):
    pass


class SDXLAutoencoderSliced(SlicedEncoderMixin, SDXLAutoencoder):
    pass


def add_sliced_attention_to_scaled_dot_product_attention(cls):
    """
    Patch refiners ScaledDotProductAttention so that it uses sliced attention

    It reduces peak memory usage.
    """

    def _sliced_attention(self, query, key, value, slice_size, is_causal=None):
        _, num_queries, _ = query.shape
        output = torch.zeros_like(query)
        for start_idx in range(0, num_queries, slice_size):
            end_idx = min(start_idx + slice_size, num_queries)
            output[:, start_idx:end_idx, :] = self._process_attention(
                query[:, start_idx:end_idx, :], key, value, is_causal
            )
        return output

    cls._sliced_attention = _sliced_attention

    def new_forward(self, query, key, value, is_causal=None):
        return self._sliced_attention(
            query, key, value, is_causal=is_causal, slice_size=2048
        )

    cls.forward = new_forward

    def _process_attention(self, query, key, value, is_causal=None):
        return self.merge_multi_head(
            x=self.dot_product(
                query=self.split_to_multi_head(query),
                key=self.split_to_multi_head(key),
                value=self.split_to_multi_head(value),
                is_causal=(
                    is_causal
                    if is_causal is not None
                    else (self.is_causal if self.is_causal is not None else False)
                ),
            )
        )

    cls._process_attention = _process_attention
    logger.debug(f"Patched {cls.__name__} with sliced attention")


add_sliced_attention_to_scaled_dot_product_attention(ScaledDotProductAttention)


@lru_cache
def monkeypatch_sd1controlnetadapter():
    """
    Another horrible thing.

    I needed to be able to cache the controlnet objects so I wouldn't be making new ones on every image generation.
    """

    def __init__(
        self,
        target: SD1UNet,
        name: str,
        weights_location: str,
    ) -> None:
        self.name = name
        controlnet = get_controlnet(
            name=name,
            weights_location=weights_location,
            device=target.device,
            dtype=target.dtype,
        )
        logger.debug(
            f"controlnet: {name} loaded to device {target.device} and type {target.dtype}"
        )

        self._controlnet: list[Controlnet] = [  # type: ignore
            controlnet
        ]  # not registered by PyTorch

        with self.setup_adapter(target):
            super(SD1ControlnetAdapter, self).__init__(target)

    SD1ControlnetAdapter.__init__ = __init__


monkeypatch_sd1controlnetadapter()


@lru_cache
def monkeypatch_self_attention_guidance():
    def new_compute_attention_scores(
        self, query: Tensor, key: Tensor, value: Tensor, slice_size=2048
    ) -> Tensor:
        query, key = self.split_to_multi_head(query), self.split_to_multi_head(key)
        batch_size, num_heads, num_queries, dim = query.shape

        output = torch.zeros(
            batch_size,
            num_heads,
            num_queries,
            num_queries,
            device=query.device,
            dtype=query.dtype,
        )
        for start_idx in range(0, num_queries, slice_size):
            end_idx = min(start_idx + slice_size, num_queries)
            sliced_query = query[:, :, start_idx:end_idx, :]
            attention_slice = sliced_query @ key.permute(0, 1, 3, 2)
            attention_slice = attention_slice / math.sqrt(dim)
            output_slice = torch.softmax(input=attention_slice, dim=-1)
            output[:, :, start_idx:end_idx, :] = output_slice

        return output

    def compute_attention_scores_logged(
        self, query: Tensor, key: Tensor, value: Tensor
    ) -> Tensor:
        print(f"query.shape: {query.shape}")
        print(f"key.shape: {key.shape}")
        query, key = self.split_to_multi_head(query), self.split_to_multi_head(key)
        print(f"mh query.shape: {query.shape}")
        print(f"mh key.shape: {key.shape}")
        _, _, _, dim = query.shape
        attention = query @ key.permute(0, 1, 3, 2)
        attention = attention / math.sqrt(dim)
        print(f"attention shape: {attention.shape}")
        result = torch.softmax(input=attention, dim=-1)
        print(f"result.shape: {result.shape}")
        return result

    SelfAttentionMap.compute_attention_scores = new_compute_attention_scores


monkeypatch_self_attention_guidance()


@lru_cache(maxsize=1)
def get_controlnet(name, weights_location, device, dtype):
    from imaginairy.utils.model_manager import load_state_dict

    controlnet_state_dict = load_state_dict(weights_location, half_mode=False)
    controlnet_state_dict = cast_weights(
        source_weights=controlnet_state_dict,
        source_model_name="controlnet-1-1",
        source_component_name="all",
        source_format="diffusers",
        dest_format="refiners",
    )

    controlnet = Controlnet(name=name, scale=1, device=device, dtype=dtype)
    controlnet.load_state_dict(controlnet_state_dict)
    return controlnet



"""Classes for discriminator loss computation"""

from typing import Dict, Iterator, List, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
import torchvision
from einops import rearrange

from imaginairy.modules.sgm.autoencoding.lpips.loss.lpips import LPIPS
from imaginairy.modules.sgm.autoencoding.lpips.model.model import weights_init
from imaginairy.modules.sgm.autoencoding.lpips.vqperceptual import (
    hinge_d_loss,
    vanilla_d_loss,
)
from imaginairy.utils import default, instantiate_from_config


class GeneralLPIPSWithDiscriminator(nn.Module):
    def __init__(
        self,
        disc_start: int,
        logvar_init: float = 0.0,
        disc_num_layers: int = 3,
        disc_in_channels: int = 3,
        disc_factor: float = 1.0,
        disc_weight: float = 1.0,
        perceptual_weight: float = 1.0,
        disc_loss: str = "hinge",
        scale_input_to_tgt_size: bool = False,
        dims: int = 2,
        learn_logvar: bool = False,
        regularization_weights: Union[None, Dict[str, float]] = None,
        additional_log_keys: Optional[List[str]] = None,
        discriminator_config: Optional[Dict] = None,
    ):
        super().__init__()
        self.dims = dims
        if self.dims > 2:
            print(
                f"running with dims={dims}. This means that for perceptual loss "
                f"calculation, the LPIPS loss will be applied to each frame "
                f"independently."
            )
        self.scale_input_to_tgt_size = scale_input_to_tgt_size
        assert disc_loss in ["hinge", "vanilla"]
        self.perceptual_loss = LPIPS().eval()
        self.perceptual_weight = perceptual_weight
        # output log variance
        self.logvar = nn.Parameter(
            torch.full((), logvar_init), requires_grad=learn_logvar
        )
        self.learn_logvar = learn_logvar

        discriminator_config = default(
            discriminator_config,
            {
                "target": "sgm.modules.autoencoding.lpips.model.model.NLayerDiscriminator",
                "params": {
                    "input_nc": disc_in_channels,
                    "n_layers": disc_num_layers,
                    "use_actnorm": False,
                },
            },
        )

        self.discriminator = instantiate_from_config(discriminator_config).apply(
            weights_init
        )
        self.discriminator_iter_start = disc_start
        self.disc_loss = hinge_d_loss if disc_loss == "hinge" else vanilla_d_loss
        self.disc_factor = disc_factor
        self.discriminator_weight = disc_weight
        self.regularization_weights = default(regularization_weights, {})

        self.forward_keys = [
            "optimizer_idx",
            "global_step",
            "last_layer",
            "split",
            "regularization_log",
        ]

        self.additional_log_keys = set(default(additional_log_keys, []))
        self.additional_log_keys.update(set(self.regularization_weights.keys()))

    def get_trainable_parameters(self) -> Iterator[nn.Parameter]:
        return self.discriminator.parameters()

    def get_trainable_autoencoder_parameters(self) -> Iterator[nn.Parameter]:
        if self.learn_logvar:
            yield self.logvar
        yield from ()

    @torch.no_grad()
    def log_images(
        self, inputs: torch.Tensor, reconstructions: torch.Tensor
    ) -> Dict[str, torch.Tensor]:
        from matplotlib import colormaps, pyplot as plt

        # calc logits of real/fake
        logits_real = self.discriminator(inputs.contiguous().detach())
        if len(logits_real.shape) < 4:
            # Non patch-discriminator
            return {}
        logits_fake = self.discriminator(reconstructions.contiguous().detach())
        # -> (b, 1, h, w)

        # parameters for colormapping
        high = max(logits_fake.abs().max(), logits_real.abs().max()).item()
        cmap = colormaps["PiYG"]  # diverging colormap

        def to_colormap(logits: torch.Tensor) -> torch.Tensor:
            """(b, 1, ...) -> (b, 3, ...)"""
            logits = (logits + high) / (2 * high)
            logits_np = cmap(logits.cpu().numpy())[..., :3]  # truncate alpha channel
            # -> (b, 1, ..., 3)
            logits = torch.from_numpy(logits_np).to(logits.device)
            return rearrange(logits, "b 1 ... c -> b c ...")

        logits_real = torch.nn.functional.interpolate(
            logits_real,
            size=inputs.shape[-2:],
            mode="nearest",
            antialias=False,
        )
        logits_fake = torch.nn.functional.interpolate(
            logits_fake,
            size=reconstructions.shape[-2:],
            mode="nearest",
            antialias=False,
        )

        # alpha value of logits for overlay
        alpha_real = torch.abs(logits_real) / high
        alpha_fake = torch.abs(logits_fake) / high
        # -> (b, 1, h, w) in range [0, 0.5]
        # alpha value of lines don't really matter, since the values are the same
        # for both images and logits anyway
        grid_alpha_real = torchvision.utils.make_grid(alpha_real, nrow=4)
        grid_alpha_fake = torchvision.utils.make_grid(alpha_fake, nrow=4)
        grid_alpha = 0.8 * torch.cat((grid_alpha_real, grid_alpha_fake), dim=1)
        # -> (1, h, w)
        # blend logits and images together

        # prepare logits for plotting
        logits_real = to_colormap(logits_real)
        logits_fake = to_colormap(logits_fake)
        # resize logits
        # -> (b, 3, h, w)

        # make some grids
        # add all logits to one plot
        logits_real = torchvision.utils.make_grid(logits_real, nrow=4)
        logits_fake = torchvision.utils.make_grid(logits_fake, nrow=4)
        # I just love how torchvision calls the number of columns `nrow`
        grid_logits = torch.cat((logits_real, logits_fake), dim=1)
        # -> (3, h, w)

        grid_images_real = torchvision.utils.make_grid(0.5 * inputs + 0.5, nrow=4)
        grid_images_fake = torchvision.utils.make_grid(
            0.5 * reconstructions + 0.5, nrow=4
        )
        grid_images = torch.cat((grid_images_real, grid_images_fake), dim=1)
        # -> (3, h, w) in range [0, 1]

        grid_blend = grid_alpha * grid_logits + (1 - grid_alpha) * grid_images

        # Create labeled colorbar
        dpi = 100
        height = 128 / dpi
        width = grid_logits.shape[2] / dpi
        fig, ax = plt.subplots(figsize=(width, height), dpi=dpi)
        img = ax.imshow(np.array([[-high, high]]), cmap=cmap)
        plt.colorbar(
            img,
            cax=ax,
            orientation="horizontal",
            fraction=0.9,
            aspect=width / height,
            pad=0.0,
        )
        img.set_visible(False)
        fig.tight_layout()
        fig.canvas.draw()
        # manually convert figure to numpy
        cbar_np = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
        cbar_np = cbar_np.reshape(fig.canvas.get_width_height()[::-1] + (3,))
        cbar = torch.from_numpy(cbar_np.copy()).to(grid_logits.dtype) / 255.0
        cbar = rearrange(cbar, "h w c -> c h w").to(grid_logits.device)

        # Add colorbar to plot
        annotated_grid = torch.cat((grid_logits, cbar), dim=1)
        blended_grid = torch.cat((grid_blend, cbar), dim=1)
        return {
            "vis_logits": 2 * annotated_grid[None, ...] - 1,
            "vis_logits_blended": 2 * blended_grid[None, ...] - 1,
        }

    def calculate_adaptive_weight(
        self, nll_loss: torch.Tensor, g_loss: torch.Tensor, last_layer: torch.Tensor
    ) -> torch.Tensor:
        nll_grads = torch.autograd.grad(nll_loss, last_layer, retain_graph=True)[0]
        g_grads = torch.autograd.grad(g_loss, last_layer, retain_graph=True)[0]

        d_weight = torch.norm(nll_grads) / (torch.norm(g_grads) + 1e-4)
        d_weight = torch.clamp(d_weight, 0.0, 1e4).detach()
        d_weight = d_weight * self.discriminator_weight
        return d_weight

    def forward(
        self,
        inputs: torch.Tensor,
        reconstructions: torch.Tensor,
        *,  # added because I changed the order here
        regularization_log: Dict[str, torch.Tensor],
        optimizer_idx: int,
        global_step: int,
        last_layer: torch.Tensor,
        split: str = "train",
        weights: Union[None, float, torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, dict]:
        if self.scale_input_to_tgt_size:
            inputs = torch.nn.functional.interpolate(
                inputs, reconstructions.shape[2:], mode="bicubic", antialias=True
            )

        if self.dims > 2:
            inputs, reconstructions = (
                rearrange(x, "b c t h w -> (b t) c h w")
                for x in (inputs, reconstructions)
            )

        rec_loss = torch.abs(inputs.contiguous() - reconstructions.contiguous())
        if self.perceptual_weight > 0:
            p_loss = self.perceptual_loss(
                inputs.contiguous(), reconstructions.contiguous()
            )
            rec_loss = rec_loss + self.perceptual_weight * p_loss

        nll_loss, weighted_nll_loss = self.get_nll_loss(rec_loss, weights)

        # now the GAN part
        if optimizer_idx == 0:
            # generator update
            if global_step >= self.discriminator_iter_start or not self.training:
                logits_fake = self.discriminator(reconstructions.contiguous())
                g_loss = -torch.mean(logits_fake)
                if self.training:
                    d_weight = self.calculate_adaptive_weight(
                        nll_loss, g_loss, last_layer=last_layer
                    )
                else:
                    d_weight = torch.tensor(1.0)
            else:
                d_weight = torch.tensor(0.0)
                g_loss = torch.tensor(0.0, requires_grad=True)

            loss = weighted_nll_loss + d_weight * self.disc_factor * g_loss
            log = {}
            for k in regularization_log:
                if k in self.regularization_weights:
                    loss = loss + self.regularization_weights[k] * regularization_log[k]
                if k in self.additional_log_keys:
                    log[f"{split}/{k}"] = regularization_log[k].detach().float().mean()

            log.update(
                {
                    f"{split}/loss/total": loss.clone().detach().mean(),
                    f"{split}/loss/nll": nll_loss.detach().mean(),
                    f"{split}/loss/rec": rec_loss.detach().mean(),
                    f"{split}/loss/g": g_loss.detach().mean(),
                    f"{split}/scalars/logvar": self.logvar.detach(),
                    f"{split}/scalars/d_weight": d_weight.detach(),
                }
            )

            return loss, log
        elif optimizer_idx == 1:
            # second pass for discriminator update
            logits_real = self.discriminator(inputs.contiguous().detach())
            logits_fake = self.discriminator(reconstructions.contiguous().detach())

            if global_step >= self.discriminator_iter_start or not self.training:
                d_loss = self.disc_factor * self.disc_loss(logits_real, logits_fake)
            else:
                d_loss = torch.tensor(0.0, requires_grad=True)

            log = {
                f"{split}/loss/disc": d_loss.clone().detach().mean(),
                f"{split}/logits/real": logits_real.detach().mean(),
                f"{split}/logits/fake": logits_fake.detach().mean(),
            }
            return d_loss, log
        else:
            msg = f"Unknown optimizer_idx {optimizer_idx}"
            raise NotImplementedError(msg)

    def get_nll_loss(
        self,
        rec_loss: torch.Tensor,
        weights: Optional[Union[float, torch.Tensor]] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        nll_loss = rec_loss / torch.exp(self.logvar) + self.logvar
        weighted_nll_loss = nll_loss
        if weights is not None:
            weighted_nll_loss = weights * nll_loss
        weighted_nll_loss = torch.sum(weighted_nll_loss) / weighted_nll_loss.shape[0]
        nll_loss = torch.sum(nll_loss) / nll_loss.shape[0]

        return nll_loss, weighted_nll_loss



import math

import torch
from jaxtyping import Float
from torch import Tensor, device as Device, dtype as DType
from torch.nn.functional import scaled_dot_product_attention as _scaled_dot_product_attention  # type: ignore

from imaginairy.vendored.refiners.fluxion.context import Contexts
from imaginairy.vendored.refiners.fluxion.layers.basics import Identity
from imaginairy.vendored.refiners.fluxion.layers.chain import Chain, Distribute, Lambda, Parallel
from imaginairy.vendored.refiners.fluxion.layers.linear import Linear
from imaginairy.vendored.refiners.fluxion.layers.module import Module


def scaled_dot_product_attention(
    query: Float[Tensor, "batch source_sequence_length dim"],
    key: Float[Tensor, "batch target_sequence_length dim"],
    value: Float[Tensor, "batch target_sequence_length dim"],
    is_causal: bool = False,
) -> Float[Tensor, "batch source_sequence_length dim"]:
    return _scaled_dot_product_attention(query, key, value, is_causal=is_causal)  # type: ignore


def sparse_dot_product_attention_non_optimized(
    query: Float[Tensor, "batch source_sequence_length dim"],
    key: Float[Tensor, "batch target_sequence_length dim"],
    value: Float[Tensor, "batch target_sequence_length dim"],
    is_causal: bool = False,
) -> Float[Tensor, "batch source_sequence_length dim"]:
    if is_causal:
        # TODO: implement causal attention
        raise NotImplementedError("Causal attention for non_optimized attention is not yet implemented")
    _, _, _, dim = query.shape
    attention = query @ key.permute(0, 1, 3, 2)
    attention = attention / math.sqrt(dim)
    attention = torch.softmax(input=attention, dim=-1)
    return attention @ value


class ScaledDotProductAttention(Module):
    def __init__(
        self,
        num_heads: int = 1,
        is_causal: bool | None = None,
        is_optimized: bool = True,
        slice_size: int | None = None,
    ) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.is_causal = is_causal
        self.is_optimized = is_optimized
        self.slice_size = slice_size
        self.dot_product = (
            scaled_dot_product_attention if self.is_optimized else sparse_dot_product_attention_non_optimized
        )

    def forward(
        self,
        query: Float[Tensor, "batch num_queries embedding_dim"],
        key: Float[Tensor, "batch num_keys embedding_dim"],
        value: Float[Tensor, "batch num_values embedding_dim"],
        is_causal: bool | None = None,
    ) -> Float[Tensor, "batch num_queries dim"]:
        if self.slice_size is None:
            return self._process_attention(query, key, value, is_causal)

        return self._sliced_attention(query, key, value, is_causal=is_causal, slice_size=self.slice_size)

    def _sliced_attention(
        self,
        query: Float[Tensor, "batch num_queries embedding_dim"],
        key: Float[Tensor, "batch num_keys embedding_dim"],
        value: Float[Tensor, "batch num_values embedding_dim"],
        slice_size: int,
        is_causal: bool | None = None,
    ) -> Float[Tensor, "batch num_queries dim"]:
        _, num_queries, _ = query.shape
        output = torch.zeros_like(query)
        for start_idx in range(0, num_queries, slice_size):
            end_idx = min(start_idx + slice_size, num_queries)
            output[:, start_idx:end_idx, :] = self._process_attention(
                query[:, start_idx:end_idx, :], key, value, is_causal
            )
        return output

    def _process_attention(
        self,
        query: Float[Tensor, "batch num_queries embedding_dim"],
        key: Float[Tensor, "batch num_keys embedding_dim"],
        value: Float[Tensor, "batch num_values embedding_dim"],
        is_causal: bool | None = None,
    ) -> Float[Tensor, "batch num_queries dim"]:
        return self.merge_multi_head(
            x=self.dot_product(
                query=self.split_to_multi_head(query),
                key=self.split_to_multi_head(key),
                value=self.split_to_multi_head(value),
                is_causal=(
                    is_causal if is_causal is not None else (self.is_causal if self.is_causal is not None else False)
                ),
            )
        )

    def split_to_multi_head(
        self, x: Float[Tensor, "batch_size sequence_length embedding_dim"]
    ) -> Float[Tensor, "batch_size num_heads sequence_length (embedding_dim//num_heads)"]:
        assert (
            len(x.shape) == 3
        ), f"Expected tensor with shape (batch_size sequence_length embedding_dim), got {x.shape}"
        assert (
            x.shape[-1] % self.num_heads == 0
        ), f"Embedding dim (x.shape[-1]={x.shape[-1]}) must be divisible by num heads"
        return x.reshape(x.shape[0], x.shape[1], self.num_heads, x.shape[-1] // self.num_heads).transpose(1, 2)

    def merge_multi_head(
        self, x: Float[Tensor, "batch_size num_heads sequence_length heads_dim"]
    ) -> Float[Tensor, "batch_size sequence_length heads_dim * num_heads"]:
        return x.transpose(1, 2).reshape(x.shape[0], x.shape[2], self.num_heads * x.shape[-1])


class Attention(Chain):
    def __init__(
        self,
        embedding_dim: int,
        num_heads: int = 1,
        key_embedding_dim: int | None = None,
        value_embedding_dim: int | None = None,
        inner_dim: int | None = None,
        use_bias: bool = True,
        is_causal: bool | None = None,
        is_optimized: bool = True,
        device: Device | str | None = None,
        dtype: DType | None = None,
    ) -> None:
        assert (
            embedding_dim % num_heads == 0
        ), f"embedding_dim {embedding_dim} must be divisible by num_heads {num_heads}"
        self.embedding_dim = embedding_dim
        self.num_heads = num_heads
        self.heads_dim = embedding_dim // num_heads
        self.key_embedding_dim = key_embedding_dim or embedding_dim
        self.value_embedding_dim = value_embedding_dim or embedding_dim
        self.inner_dim = inner_dim or embedding_dim
        self.use_bias = use_bias
        self.is_causal = is_causal
        self.is_optimized = is_optimized
        super().__init__(
            Distribute(
                Linear(
                    in_features=self.embedding_dim,
                    out_features=self.inner_dim,
                    bias=self.use_bias,
                    device=device,
                    dtype=dtype,
                ),
                Linear(
                    in_features=self.key_embedding_dim,
                    out_features=self.inner_dim,
                    bias=self.use_bias,
                    device=device,
                    dtype=dtype,
                ),
                Linear(
                    in_features=self.value_embedding_dim,
                    out_features=self.inner_dim,
                    bias=self.use_bias,
                    device=device,
                    dtype=dtype,
                ),
            ),
            ScaledDotProductAttention(num_heads=num_heads, is_causal=is_causal, is_optimized=is_optimized),
            Linear(
                in_features=self.inner_dim,
                out_features=self.embedding_dim,
                bias=True,
                device=device,
                dtype=dtype,
            ),
        )


class SelfAttention(Attention):
    def __init__(
        self,
        embedding_dim: int,
        inner_dim: int | None = None,
        num_heads: int = 1,
        use_bias: bool = True,
        is_causal: bool | None = None,
        is_optimized: bool = True,
        device: Device | str | None = None,
        dtype: DType | None = None,
    ) -> None:
        super().__init__(
            embedding_dim=embedding_dim,
            inner_dim=inner_dim,
            num_heads=num_heads,
            use_bias=use_bias,
            is_causal=is_causal,
            is_optimized=is_optimized,
            device=device,
            dtype=dtype,
        )
        self.insert(0, Parallel(Identity(), Identity(), Identity()))


class SelfAttention2d(SelfAttention):
    def __init__(
        self,
        channels: int,
        num_heads: int = 1,
        use_bias: bool = True,
        is_causal: bool | None = None,
        is_optimized: bool = True,
        device: Device | str | None = None,
        dtype: DType | None = None,
    ) -> None:
        assert channels % num_heads == 0, f"channels {channels} must be divisible by num_heads {num_heads}"
        self.channels = channels
        super().__init__(
            embedding_dim=channels,
            num_heads=num_heads,
            use_bias=use_bias,
            is_causal=is_causal,
            is_optimized=is_optimized,
            device=device,
            dtype=dtype,
        )
        self.insert(0, Lambda(self.tensor_2d_to_sequence))
        self.append(Lambda(self.sequence_to_tensor_2d))

    def init_context(self) -> Contexts:
        return {"reshape": {"height": None, "width": None}}

    def tensor_2d_to_sequence(
        self, x: Float[Tensor, "batch channels height width"]
    ) -> Float[Tensor, "batch height*width channels"]:
        height, width = x.shape[-2:]
        self.set_context(context="reshape", value={"height": height, "width": width})
        return x.reshape(x.shape[0], x.shape[1], height * width).transpose(1, 2)

    def sequence_to_tensor_2d(
        self, x: Float[Tensor, "batch sequence_length channels"]
    ) -> Float[Tensor, "batch channels height width"]:
        height, width = self.use_context("reshape").values()
        return x.transpose(1, 2).reshape(x.shape[0], x.shape[2], height, width)



import types
from typing import Optional

import numpy as np
import timm
import torch
import torch.nn.functional as F
from timm.models.beit import gen_relative_position_index
from torch.utils.checkpoint import checkpoint

from .utils import forward_adapted_unflatten, make_backbone_default


def forward_beit(pretrained, x):
    return forward_adapted_unflatten(pretrained, x, "forward_features")


def patch_embed_forward(self, x):
    """
    Modification of timm.models.layers.patch_embed.py: PatchEmbed.forward to support arbitrary window sizes.
    """
    x = self.proj(x)
    if self.flatten:
        x = x.flatten(2).transpose(1, 2)
    x = self.norm(x)
    return x


def _get_rel_pos_bias(self, window_size):
    """
    Modification of timm.models.beit.py: Attention._get_rel_pos_bias to support arbitrary window sizes.
    """
    old_height = 2 * self.window_size[0] - 1
    old_width = 2 * self.window_size[1] - 1

    new_height = 2 * window_size[0] - 1
    new_width = 2 * window_size[1] - 1

    old_relative_position_bias_table = self.relative_position_bias_table

    old_num_relative_distance = self.num_relative_distance
    new_num_relative_distance = new_height * new_width + 3

    old_sub_table = old_relative_position_bias_table[: old_num_relative_distance - 3]

    old_sub_table = old_sub_table.reshape(1, old_width, old_height, -1).permute(
        0, 3, 1, 2
    )
    new_sub_table = F.interpolate(
        old_sub_table, size=(int(new_height), int(new_width)), mode="bilinear"
    )
    new_sub_table = new_sub_table.permute(0, 2, 3, 1).reshape(
        new_num_relative_distance - 3, -1
    )

    new_relative_position_bias_table = torch.cat(
        [
            new_sub_table,
            old_relative_position_bias_table[old_num_relative_distance - 3 :],
        ]
    )

    key = str(window_size[1]) + "," + str(window_size[0])
    if key not in self.relative_position_indices:
        self.relative_position_indices[key] = gen_relative_position_index(window_size)

    relative_position_bias = new_relative_position_bias_table[
        self.relative_position_indices[key].view(-1)
    ].view(
        window_size[0] * window_size[1] + 1, window_size[0] * window_size[1] + 1, -1
    )  # Wh*Ww,Wh*Ww,nH
    relative_position_bias = relative_position_bias.permute(
        2, 0, 1
    ).contiguous()  # nH, Wh*Ww, Wh*Ww
    return relative_position_bias.unsqueeze(0)


def attention_forward(
    self, x, resolution, shared_rel_pos_bias: Optional[torch.Tensor] = None
):
    """
    Modification of timm.models.beit.py: Attention.forward to support arbitrary window sizes.
    """
    B, N, C = x.shape

    qkv_bias = (
        torch.cat((self.q_bias, self.k_bias, self.v_bias))
        if self.q_bias is not None
        else None
    )
    qkv = F.linear(input=x, weight=self.qkv.weight, bias=qkv_bias)
    qkv = qkv.reshape(B, N, 3, self.num_heads, -1).permute(2, 0, 3, 1, 4)
    q, k, v = qkv.unbind(0)  # make torchscript happy (cannot use tensor as tuple)

    q = q * self.scale
    attn = q @ k.transpose(-2, -1)

    if self.relative_position_bias_table is not None:
        window_size = tuple(np.array(resolution) // 16)
        attn = attn + self._get_rel_pos_bias(window_size)
    if shared_rel_pos_bias is not None:
        attn = attn + shared_rel_pos_bias

    attn = attn.softmax(dim=-1)
    attn = self.attn_drop(attn)

    x = (attn @ v).transpose(1, 2).reshape(B, N, -1)
    x = self.proj(x)
    x = self.proj_drop(x)
    return x


def block_forward(
    self, x, resolution, shared_rel_pos_bias: Optional[torch.Tensor] = None
):
    """
    Modification of timm.models.beit.py: Block.forward to support arbitrary window sizes.
    See https://github.com/isl-org/MiDaS/pull/234
    """
    if self.gamma_1 is None:
        x = x + self.drop_path1(
            self.attn(
                self.norm1(x), resolution, shared_rel_pos_bias=shared_rel_pos_bias
            )
        )
        x = x + self.drop_path2(self.mlp(self.norm2(x)))
    else:
        x = x + self.drop_path1(
            self.gamma_1
            * self.attn(
                self.norm1(x), resolution, shared_rel_pos_bias=shared_rel_pos_bias
            )
        )
        x = x + self.drop_path2(self.gamma_2 * self.mlp(self.norm2(x)))
    return x


def beit_forward_features(self, x):
    """
    Modification of timm.models.beit.py: Beit.forward_features to support arbitrary window sizes.
    """
    resolution = x.shape[2:]

    x = self.patch_embed(x)
    x = torch.cat((self.cls_token.expand(x.shape[0], -1, -1), x), dim=1)
    if self.pos_embed is not None:
        x = x + self.pos_embed
    x = self.pos_drop(x)

    rel_pos_bias = self.rel_pos_bias() if self.rel_pos_bias is not None else None
    for blk in self.blocks:
        if self.grad_checkpointing and not torch.jit.is_scripting():
            x = checkpoint(blk, x, shared_rel_pos_bias=rel_pos_bias)
        else:
            x = blk(x, resolution, shared_rel_pos_bias=rel_pos_bias)
    x = self.norm(x)
    return x


def _make_beit_backbone(
    model,
    features=[96, 192, 384, 768],
    size=[384, 384],
    hooks=[0, 4, 8, 11],
    vit_features=768,
    use_readout="ignore",
    start_index=1,
    start_index_readout=1,
):
    backbone = make_backbone_default(
        model,
        features,
        size,
        hooks,
        vit_features,
        use_readout,
        start_index,
        start_index_readout,
    )

    backbone.model.patch_embed.forward = types.MethodType(
        patch_embed_forward, backbone.model.patch_embed
    )
    backbone.model.forward_features = types.MethodType(
        beit_forward_features, backbone.model
    )

    for block in backbone.model.blocks:
        attn = block.attn
        attn._get_rel_pos_bias = types.MethodType(_get_rel_pos_bias, attn)
        attn.forward = types.MethodType(attention_forward, attn)
        attn.relative_position_indices = {}

        block.forward = types.MethodType(block_forward, block)

    return backbone


def _make_pretrained_beitl16_512(pretrained, use_readout="ignore", hooks=None):
    model = timm.create_model("beit_large_patch16_512", pretrained=pretrained)

    hooks = [5, 11, 17, 23] if hooks is None else hooks

    features = [256, 512, 1024, 1024]

    return _make_beit_backbone(
        model,
        features=features,
        size=[512, 512],
        hooks=hooks,
        vit_features=1024,
        use_readout=use_readout,
    )


def _make_pretrained_beitl16_384(pretrained, use_readout="ignore", hooks=None):
    model = timm.create_model("beit_large_patch16_384", pretrained=pretrained)

    hooks = [5, 11, 17, 23] if hooks is None else hooks
    return _make_beit_backbone(
        model,
        features=[256, 512, 1024, 1024],
        hooks=hooks,
        vit_features=1024,
        use_readout=use_readout,
    )


def _make_pretrained_beitb16_384(pretrained, use_readout="ignore", hooks=None):
    model = timm.create_model("beit_base_patch16_384", pretrained=pretrained)

    hooks = [2, 5, 8, 11] if hooks is None else hooks
    return _make_beit_backbone(
        model,
        features=[96, 192, 384, 768],
        hooks=hooks,
        use_readout=use_readout,
    )



from typing import Literal

import pyparsing
from PIL import Image, ImageDraw, ImageFont
from PIL.ImageColor import getrgb

from imaginairy.utils.img_utils import create_halo_effect
from imaginairy.utils.paths import PKG_ROOT
from imaginairy.utils.spaced_kv_parser import parse_spaced_key_value_pairs


def determine_max_font_size(
    text: str,
    draw: ImageDraw.ImageDraw,
    font_path: str,
    width: int,
    height: int,
    margin_pct: float,
    line_spacing: int = 4,
) -> int:
    """
    Determine the maximum font size that allows the text to fit within the given image dimensions and margin constraints.
    Updated to use multiline_textbbox in Pillow 10.1.0.

    :param text: Text to be drawn.
    :param draw: ImageDraw object to measure text size.
    :param font_path: Path to the font file.
    :param width: Width of the image.
    :param height: Height of the image.
    :param margin_pct: Margin percentage.
    :return: Maximum font size.
    """
    max_width = width - 2 * (width * margin_pct)
    max_height = height - 2 * (height * margin_pct)

    font_size = 1
    font = ImageFont.truetype(font_path, font_size)

    while True:
        # Use multiline_textbbox to get the bounding box of the text
        bbox = draw.multiline_textbbox((0, 0), text, font=font, spacing=line_spacing)
        text_width = bbox[2] - bbox[0]  # right - left
        text_height = bbox[3] - bbox[1]  # bottom - top

        if text_width > max_width or text_height > max_height:
            break
        font_size += 1
        font = ImageFont.truetype(font_path, font_size)

    # Subtract 1 because the loop exits after the size becomes too large
    return font_size - 1


def generate_word_image(
    text: str,
    width: int,
    height: int,
    margin_pct: float = 0.1,
    line_spacing: int = 4,
    text_align: Literal["left", "center", "right"] = "center",
    font_path: str = f"{PKG_ROOT}/data/DejaVuSans.ttf",
    font_color: str = "black",
    background_color: str = "white",
) -> Image.Image:
    image = Image.new("RGB", (width, height), color=background_color)
    draw = ImageDraw.Draw(image)

    max_font_size = determine_max_font_size(
        text, draw, font_path, width, height, margin_pct, line_spacing=line_spacing
    )

    font = ImageFont.truetype(font_path, max_font_size)

    bbox = draw.multiline_textbbox((0, 0), text, font=font)

    # Calculate text position
    text_width = bbox[2] - bbox[0]
    text_height = bbox[3] - bbox[1]
    x = (width - text_width) / 2
    y = (height - text_height) / 2 - bbox[1]

    draw.multiline_text(
        (x, y), text, fill=font_color, font=font, align=text_align, spacing=line_spacing
    )

    return image


def image_from_textimg_str(text: str, width: int, height: int) -> Image.Image:
    """
    Create an image from a textimg string.
    """
    try:
        data = parse_spaced_key_value_pairs(text)
    except pyparsing.ParseException:
        raise ValueError("Invalid format for textimg")  # noqa

    first_key = next(iter(data))

    if first_key != "textimg":
        raise ValueError("Invalid format for textimg")

    allowed_keys = {
        "textimg",
        "font",
        "font_color",
        "background_color",
        "text_align",
        "line_spacing",
        "margin_pct",
        "halo",
    }
    submitted_keys = set(data.keys())
    invalid_keys = submitted_keys - allowed_keys
    if invalid_keys:
        msg = f"Invalid attributes for textimg: '{invalid_keys}'. Valid attributes are '{allowed_keys}'"
        raise ValueError(msg)

    text_align = data.get("text_align", "center")
    valid_alignments = {"left", "center", "right"}
    if text_align not in valid_alignments:
        msg = f"Invalid text_align '{text_align}'. Valid options are 'left', 'center' and 'right'"
        raise ValueError(msg)
    assert text_align in valid_alignments
    background_color: str = data.get("background_color", "white")
    img = generate_word_image(
        text=data["textimg"].replace("\\n", "\n"),
        width=width,
        height=height,
        margin_pct=float(data.get("margin_pct", 0.1)),
        line_spacing=int(data.get("line_spacing", 4)),
        text_align=text_align,  # type: ignore
        font_path=data.get("font", f"{PKG_ROOT}/data/DejaVuSans.ttf"),
        font_color=data.get("font_color", "black"),
        background_color=background_color,
    )
    bg_color_rgb = getrgb(background_color)
    if data.get("halo", "0").lower() in ("true", "1", "yes"):
        img = create_halo_effect(img, background_color=bg_color_rgb)

    return img



"""Classes for autoencoder-based image generation"""

import logging
import math
import re
from abc import abstractmethod
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

import torch
import torch.nn as nn
from einops import rearrange
from packaging import version

from imaginairy.modules.ema import LitEma
from imaginairy.utils import (
    default,
    get_nested_attribute,
    get_obj_from_str,
    instantiate_from_config,
)

if TYPE_CHECKING:
    from .autoencoding.regularizers import AbstractRegularizer


# from .ema import LitEma
# from .util import (default, get_nested_attribute, get_obj_from_str,
#                     instantiate_from_config)

logpy = logging.getLogger(__name__)


class AbstractAutoencoder(nn.Module):
    """
    This is the base class for all autoencoders, including image autoencoders, image autoencoders with discriminators,
    unCLIP models, etc. Hence, it is fairly general, and specific features
    (e.g. discriminator training, encoding, decoding) must be implemented in subclasses.
    """

    def __init__(
        self,
        ema_decay: Union[None, float] = None,
        monitor: Union[None, str] = None,
        input_key: str = "jpg",
    ):
        super().__init__()

        self.input_key = input_key
        self.use_ema = ema_decay is not None
        if monitor is not None:
            self.monitor = monitor

        if self.use_ema:
            self.model_ema = LitEma(self, decay=ema_decay)
            logpy.info(f"Keeping EMAs of {len(list(self.model_ema.buffers()))}.")

        if version.parse(torch.__version__) >= version.parse("2.0.0"):
            self.automatic_optimization = False

    def apply_ckpt(self, ckpt: Union[None, str, dict]):
        if ckpt is None:
            return
        if isinstance(ckpt, str):
            ckpt = {
                "target": "imaginairy.modules.sgm.checkpoint.CheckpointEngine",
                "params": {"ckpt_path": ckpt},
            }
        engine = instantiate_from_config(ckpt)
        engine(self)

    @abstractmethod
    def get_input(self, batch) -> Any:
        raise NotImplementedError()

    def on_train_batch_end(self, *args, **kwargs):
        # for EMA computation
        if self.use_ema:
            self.model_ema(self)

    @contextmanager
    def ema_scope(self, context=None):
        if self.use_ema:
            self.model_ema.store(self.parameters())
            self.model_ema.copy_to(self)
            if context is not None:
                logpy.info(f"{context}: Switched to EMA weights")
        try:
            yield None
        finally:
            if self.use_ema:
                self.model_ema.restore(self.parameters())
                if context is not None:
                    logpy.info(f"{context}: Restored training weights")

    @abstractmethod
    def encode(self, *args, **kwargs) -> torch.Tensor:
        raise NotImplementedError("encode()-method of abstract base class called")

    @abstractmethod
    def decode(self, *args, **kwargs) -> torch.Tensor:
        raise NotImplementedError("decode()-method of abstract base class called")

    def instantiate_optimizer_from_config(self, params, lr, cfg):
        logpy.info(f"loading >>> {cfg['target']} <<< optimizer from config")
        return get_obj_from_str(cfg["target"])(params, lr=lr, **cfg.get("params", {}))

    def configure_optimizers(self) -> Any:
        raise NotImplementedError()


class AutoencodingEngine(AbstractAutoencoder):
    """
    Base class for all image autoencoders that we train, like VQGAN or AutoencoderKL
    (we also restore them explicitly as special cases for legacy reasons).
    Regularizations such as KL or VQ are moved to the regularizer class.
    """

    def __init__(
        self,
        *args,
        encoder_config: Dict,
        decoder_config: Dict,
        loss_config: Dict,
        regularizer_config: Dict,
        optimizer_config: Union[Dict, None] = None,
        lr_g_factor: float = 1.0,
        trainable_ae_params: Optional[List[List[str]]] = None,
        ae_optimizer_args: Optional[List[dict]] = None,
        trainable_disc_params: Optional[List[List[str]]] = None,
        disc_optimizer_args: Optional[List[dict]] = None,
        disc_start_iter: int = 0,
        diff_boost_factor: float = 3.0,
        ckpt_engine: Union[None, str, dict] = None,
        ckpt_path: Optional[str] = None,
        additional_decode_keys: Optional[List[str]] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.automatic_optimization = False  # pytorch lightning

        self.encoder: torch.nn.Module = instantiate_from_config(encoder_config)
        self.decoder: torch.nn.Module = instantiate_from_config(decoder_config)
        self.loss: torch.nn.Module = instantiate_from_config(loss_config)
        self.regularization: AbstractRegularizer = instantiate_from_config(
            regularizer_config
        )
        self.optimizer_config = default(
            optimizer_config, {"target": "torch.optim.Adam"}
        )
        self.diff_boost_factor = diff_boost_factor
        self.disc_start_iter = disc_start_iter
        self.lr_g_factor = lr_g_factor
        self.trainable_ae_params = trainable_ae_params
        if self.trainable_ae_params is not None:
            self.ae_optimizer_args = default(
                ae_optimizer_args,
                [{} for _ in range(len(self.trainable_ae_params))],
            )
            assert len(self.ae_optimizer_args) == len(self.trainable_ae_params)
        else:
            self.ae_optimizer_args = [{}]  # makes type consitent

        self.trainable_disc_params = trainable_disc_params
        if self.trainable_disc_params is not None:
            self.disc_optimizer_args = default(
                disc_optimizer_args,
                [{} for _ in range(len(self.trainable_disc_params))],
            )
            assert len(self.disc_optimizer_args) == len(self.trainable_disc_params)
        else:
            self.disc_optimizer_args = [{}]  # makes type consitent

        if ckpt_path is not None:
            assert ckpt_engine is None, "Can't set ckpt_engine and ckpt_path"
            logpy.warning(
                "Checkpoint path is deprecated, use `checkpoint_egnine` instead"
            )
        self.apply_ckpt(default(ckpt_path, ckpt_engine))
        self.additional_decode_keys = set(default(additional_decode_keys, []))

    def get_input(self, batch: Dict) -> torch.Tensor:
        # assuming unified data format, dataloader returns a dict.
        # image tensors should be scaled to -1 ... 1 and in channels-first
        # format (e.g., bchw instead if bhwc)
        return batch[self.input_key]

    def get_autoencoder_params(self) -> list:
        params = []
        if hasattr(self.loss, "get_trainable_autoencoder_parameters"):
            params += list(self.loss.get_trainable_autoencoder_parameters())
        if hasattr(self.regularization, "get_trainable_parameters"):
            params += list(self.regularization.get_trainable_parameters())
        params = params + list(self.encoder.parameters())
        params = params + list(self.decoder.parameters())
        return params

    def get_discriminator_params(self) -> list:
        if hasattr(self.loss, "get_trainable_parameters"):
            params = list(self.loss.get_trainable_parameters())  # e.g., discriminator
        else:
            params = []
        return params

    def get_last_layer(self):
        return self.decoder.get_last_layer()

    def encode(
        self,
        x: torch.Tensor,
        return_reg_log: bool = False,
        unregularized: bool = False,
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, dict]]:
        z = self.encoder(x)
        if unregularized:
            return z, {}
        z, reg_log = self.regularization(z)
        if return_reg_log:
            return z, reg_log
        return z

    def decode(self, z: torch.Tensor, **kwargs) -> torch.Tensor:
        x = self.decoder(z, **kwargs)
        return x

    def forward(
        self, x: torch.Tensor, **additional_decode_kwargs
    ) -> Tuple[torch.Tensor, torch.Tensor, dict]:
        z, reg_log = self.encode(x, return_reg_log=True)
        dec = self.decode(z, **additional_decode_kwargs)
        return z, dec, reg_log

    def inner_training_step(
        self, batch: dict, batch_idx: int, optimizer_idx: int = 0
    ) -> torch.Tensor:
        x = self.get_input(batch)
        additional_decode_kwargs = {
            key: batch[key] for key in self.additional_decode_keys.intersection(batch)
        }
        z, xrec, regularization_log = self(x, **additional_decode_kwargs)
        if hasattr(self.loss, "forward_keys"):
            extra_info = {
                "z": z,
                "optimizer_idx": optimizer_idx,
                "global_step": self.global_step,
                "last_layer": self.get_last_layer(),
                "split": "train",
                "regularization_log": regularization_log,
                "autoencoder": self,
            }
            extra_info = {k: extra_info[k] for k in self.loss.forward_keys}
        else:
            extra_info = {}

        if optimizer_idx == 0:
            # autoencode
            out_loss = self.loss(x, xrec, **extra_info)
            if isinstance(out_loss, tuple):
                aeloss, log_dict_ae = out_loss
            else:
                # simple loss function
                aeloss = out_loss
                log_dict_ae = {"train/loss/rec": aeloss.detach()}

            self.log_dict(
                log_dict_ae,
                prog_bar=False,
                logger=True,
                on_step=True,
                on_epoch=True,
                sync_dist=False,
            )
            self.log(
                "loss",
                aeloss.mean().detach(),
                prog_bar=True,
                logger=False,
                on_epoch=False,
                on_step=True,
            )
            return aeloss
        elif optimizer_idx == 1:
            # discriminator
            discloss, log_dict_disc = self.loss(x, xrec, **extra_info)
            # -> discriminator always needs to return a tuple
            self.log_dict(
                log_dict_disc, prog_bar=False, logger=True, on_step=True, on_epoch=True
            )
            return discloss
        else:
            msg = f"Unknown optimizer {optimizer_idx}"
            raise NotImplementedError(msg)

    def training_step(self, batch: dict, batch_idx: int):
        opts = self.optimizers()
        if not isinstance(opts, list):
            # Non-adversarial case
            opts = [opts]
        optimizer_idx = batch_idx % len(opts)
        if self.global_step < self.disc_start_iter:
            optimizer_idx = 0
        opt = opts[optimizer_idx]
        opt.zero_grad()
        with opt.toggle_model():
            loss = self.inner_training_step(
                batch, batch_idx, optimizer_idx=optimizer_idx
            )
            self.manual_backward(loss)
        opt.step()

    def validation_step(self, batch: dict, batch_idx: int) -> Dict:
        log_dict = self._validation_step(batch, batch_idx)
        with self.ema_scope():
            log_dict_ema = self._validation_step(batch, batch_idx, postfix="_ema")
            log_dict.update(log_dict_ema)
        return log_dict

    def _validation_step(self, batch: dict, batch_idx: int, postfix: str = "") -> Dict:
        x = self.get_input(batch)

        z, xrec, regularization_log = self(x)
        if hasattr(self.loss, "forward_keys"):
            extra_info = {
                "z": z,
                "optimizer_idx": 0,
                "global_step": self.global_step,
                "last_layer": self.get_last_layer(),
                "split": "val" + postfix,
                "regularization_log": regularization_log,
                "autoencoder": self,
            }
            extra_info = {k: extra_info[k] for k in self.loss.forward_keys}
        else:
            extra_info = {}
        out_loss = self.loss(x, xrec, **extra_info)
        if isinstance(out_loss, tuple):
            aeloss, log_dict_ae = out_loss
        else:
            # simple loss function
            aeloss = out_loss
            log_dict_ae = {f"val{postfix}/loss/rec": aeloss.detach()}
        full_log_dict = log_dict_ae

        if "optimizer_idx" in extra_info:
            extra_info["optimizer_idx"] = 1
            discloss, log_dict_disc = self.loss(x, xrec, **extra_info)
            full_log_dict.update(log_dict_disc)
        self.log(
            f"val{postfix}/loss/rec",
            log_dict_ae[f"val{postfix}/loss/rec"],
            sync_dist=True,
        )
        self.log_dict(full_log_dict, sync_dist=True)
        return full_log_dict

    def get_param_groups(
        self, parameter_names: List[List[str]], optimizer_args: List[dict]
    ) -> Tuple[List[Dict[str, Any]], int]:
        groups = []
        num_params = 0
        for names, args in zip(parameter_names, optimizer_args):
            params = []
            for pattern_ in names:
                pattern_params = []
                pattern = re.compile(pattern_)
                for p_name, param in self.named_parameters():
                    if re.match(pattern, p_name):
                        pattern_params.append(param)
                        num_params += param.numel()
                if len(pattern_params) == 0:
                    logpy.warning(f"Did not find parameters for pattern {pattern_}")
                params.extend(pattern_params)
            groups.append({"params": params, **args})
        return groups, num_params

    def configure_optimizers(self) -> List[torch.optim.Optimizer]:
        if self.trainable_ae_params is None:
            ae_params = self.get_autoencoder_params()
        else:
            ae_params, num_ae_params = self.get_param_groups(
                self.trainable_ae_params, self.ae_optimizer_args
            )
            logpy.info(f"Number of trainable autoencoder parameters: {num_ae_params:,}")
        if self.trainable_disc_params is None:
            disc_params = self.get_discriminator_params()
        else:
            disc_params, num_disc_params = self.get_param_groups(
                self.trainable_disc_params, self.disc_optimizer_args
            )
            logpy.info(
                f"Number of trainable discriminator parameters: {num_disc_params:,}"
            )
        opt_ae = self.instantiate_optimizer_from_config(
            ae_params,
            default(self.lr_g_factor, 1.0) * self.learning_rate,
            self.optimizer_config,
        )
        opts = [opt_ae]
        if len(disc_params) > 0:
            opt_disc = self.instantiate_optimizer_from_config(
                disc_params, self.learning_rate, self.optimizer_config
            )
            opts.append(opt_disc)

        return opts

    @torch.no_grad()
    def log_images(
        self, batch: dict, additional_log_kwargs: Optional[Dict] = None, **kwargs
    ) -> dict:
        log = {}
        additional_decode_kwargs = {}
        x = self.get_input(batch)
        additional_decode_kwargs.update(
            {key: batch[key] for key in self.additional_decode_keys.intersection(batch)}
        )

        _, xrec, _ = self(x, **additional_decode_kwargs)
        log["inputs"] = x
        log["reconstructions"] = xrec
        diff = 0.5 * torch.abs(torch.clamp(xrec, -1.0, 1.0) - x)
        diff.clamp_(0, 1.0)
        log["diff"] = 2.0 * diff - 1.0
        # diff_boost shows location of small errors, by boosting their
        # brightness.
        log["diff_boost"] = (
            2.0 * torch.clamp(self.diff_boost_factor * diff, 0.0, 1.0) - 1
        )
        if hasattr(self.loss, "log_images"):
            log.update(self.loss.log_images(x, xrec))
        with self.ema_scope():
            _, xrec_ema, _ = self(x, **additional_decode_kwargs)
            log["reconstructions_ema"] = xrec_ema
            diff_ema = 0.5 * torch.abs(torch.clamp(xrec_ema, -1.0, 1.0) - x)
            diff_ema.clamp_(0, 1.0)
            log["diff_ema"] = 2.0 * diff_ema - 1.0
            log["diff_boost_ema"] = (
                2.0 * torch.clamp(self.diff_boost_factor * diff_ema, 0.0, 1.0) - 1
            )
        if additional_log_kwargs:
            additional_decode_kwargs.update(additional_log_kwargs)
            _, xrec_add, _ = self(x, **additional_decode_kwargs)
            log_str = "reconstructions-" + "-".join(
                [f"{key}={additional_log_kwargs[key]}" for key in additional_log_kwargs]
            )
            log[log_str] = xrec_add
        return log


class AutoencodingEngineLegacy(AutoencodingEngine):
    def __init__(self, embed_dim: int, **kwargs):
        self.max_batch_size = kwargs.pop("max_batch_size", None)
        ddconfig = kwargs.pop("ddconfig")
        ckpt_path = kwargs.pop("ckpt_path", None)
        ckpt_engine = kwargs.pop("ckpt_engine", None)
        super().__init__(
            encoder_config={
                "target": "imaginairy.modules.sgm.diffusionmodules.model.Encoder",
                "params": ddconfig,
            },
            decoder_config={
                "target": "imaginairy.modules.sgm.diffusionmodules.model.Decoder",
                "params": ddconfig,
            },
            **kwargs,
        )
        self.quant_conv = torch.nn.Conv2d(
            (1 + ddconfig["double_z"]) * ddconfig["z_channels"],
            (1 + ddconfig["double_z"]) * embed_dim,
            1,
        )
        self.post_quant_conv = torch.nn.Conv2d(embed_dim, ddconfig["z_channels"], 1)
        self.embed_dim = embed_dim

        self.apply_ckpt(default(ckpt_path, ckpt_engine))

    def get_autoencoder_params(self) -> list:
        params = super().get_autoencoder_params()
        return params

    def encode(
        self, x: torch.Tensor, return_reg_log: bool = False
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, dict]]:
        if self.max_batch_size is None:
            z = self.encoder(x)
            z = self.quant_conv(z)
        else:
            N = x.shape[0]
            bs = self.max_batch_size
            n_batches = int(math.ceil(N / bs))
            z = []
            for i_batch in range(n_batches):
                z_batch = self.encoder(x[i_batch * bs : (i_batch + 1) * bs])
                z_batch = self.quant_conv(z_batch)
                z.append(z_batch)
            z = torch.cat(z, 0)

        z, reg_log = self.regularization(z)
        if return_reg_log:
            return z, reg_log
        return z

    def decode(self, z: torch.Tensor, **decoder_kwargs) -> torch.Tensor:
        if self.max_batch_size is None:
            dec = self.post_quant_conv(z)
            dec = self.decoder(dec, **decoder_kwargs)
        else:
            N = z.shape[0]
            bs = self.max_batch_size
            n_batches = int(math.ceil(N / bs))
            dec = []
            for i_batch in range(n_batches):
                dec_batch = self.post_quant_conv(z[i_batch * bs : (i_batch + 1) * bs])
                dec_batch = self.decoder(dec_batch, **decoder_kwargs)
                dec.append(dec_batch)
            dec = torch.cat(dec, 0)

        return dec


class AutoencoderKL(AutoencodingEngineLegacy):
    def __init__(self, **kwargs):
        if "lossconfig" in kwargs:
            kwargs["loss_config"] = kwargs.pop("lossconfig")
        super().__init__(
            regularizer_config={
                "target": (
                    "imaginairy.modules.sgm.autoencoding.regularizers"
                    ".DiagonalGaussianRegularizer"
                )
            },
            **kwargs,
        )


class AutoencoderLegacyVQ(AutoencodingEngineLegacy):
    def __init__(
        self,
        embed_dim: int,
        n_embed: int,
        sane_index_shape: bool = False,
        **kwargs,
    ):
        if "lossconfig" in kwargs:
            logpy.warning("Parameter `lossconfig` is deprecated, use `loss_config`.")
            kwargs["loss_config"] = kwargs.pop("lossconfig")
        super().__init__(
            regularizer_config={
                "target": (
                    "imaginairy.modules.sgm.autoencoding.regularizers.quantize"
                    ".VectorQuantizer"
                ),
                "params": {
                    "n_e": n_embed,
                    "e_dim": embed_dim,
                    "sane_index_shape": sane_index_shape,
                },
            },
            **kwargs,
        )


class IdentityFirstStage(AbstractAutoencoder):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_input(self, x: Any) -> Any:
        return x

    def encode(self, x: Any, *args, **kwargs) -> Any:
        return x

    def decode(self, x: Any, *args, **kwargs) -> Any:
        return x


class AEIntegerWrapper(nn.Module):
    def __init__(
        self,
        model: nn.Module,
        shape: Union[None, Tuple[int, int], List[int]] = (16, 16),
        regularization_key: str = "regularization",
        encoder_kwargs: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        self.model = model
        if not hasattr(model, "encode") or hasattr(model, "decode"):
            raise RuntimeError("Need AE interface")
        self.regularization = get_nested_attribute(model, regularization_key)
        self.shape = shape
        self.encoder_kwargs = default(encoder_kwargs, {"return_reg_log": True})

    def encode(self, x) -> torch.Tensor:
        assert (
            not self.training
        ), f"{self.__class__.__name__} only supports inference currently"
        _, log = self.model.encode(x, **self.encoder_kwargs)
        assert isinstance(log, dict)
        inds = log["min_encoding_indices"]
        return rearrange(inds, "b ... -> b (...)")

    def decode(
        self, inds: torch.Tensor, shape: Union[None, tuple, list] = None
    ) -> torch.Tensor:
        # expect inds shape (b, s) with s = h*w
        shape = default(shape, self.shape)  # Optional[(h, w)]
        if shape is not None:
            assert len(shape) == 2, f"Unhandeled shape {shape}"
            inds = rearrange(inds, "b (h w) -> b h w", h=shape[0], w=shape[1])
        h = self.regularization.get_codebook_entry(inds)  # (b, h, w, c)
        h = rearrange(h, "b h w c -> b c h w")
        return self.model.decode(h)


class AutoencoderKLModeOnly(AutoencodingEngineLegacy):
    def __init__(self, **kwargs):
        if "lossconfig" in kwargs:
            kwargs["loss_config"] = kwargs.pop("lossconfig")
        super().__init__(
            regularizer_config={
                "target": (
                    "imaginairy.modules.sgm.autoencoding.regularizers"
                    ".DiagonalGaussianRegularizer"
                ),
                "params": {"sample": False},
            },
            **kwargs,
        )



import pytest

from imaginairy.api import imagine
from imaginairy.schema import ImaginePrompt, LazyLoadingImage
from imaginairy.utils import get_device
from imaginairy.utils.outpaint import outpaint_arg_str_parse
from tests import TESTS_FOLDER
from tests.utils import assert_image_similar_to_expectation


@pytest.mark.skipif(get_device() == "cpu", reason="Too slow to run on CPU")
def test_outpainting_outpaint(filename_base_for_outputs):
    img = LazyLoadingImage(
        filepath=f"{TESTS_FOLDER}/data/girl_with_a_pearl_earring.jpg"
    )
    prompt = ImaginePrompt(
        prompt="woman standing",
        init_image=img,
        init_image_strength=0,
        mask_prompt="background",
        outpaint="all250,up0,down600",
        mask_mode="replace",
        negative_prompt="picture frame, borders, framing, text, writing, watermarks, indoors, advertisement, paper, canvas, stock photo",
        steps=20,
        seed=542906833,
    )
    result = next(iter(imagine([prompt])))
    img_path = f"{filename_base_for_outputs}.png"
    assert_image_similar_to_expectation(result.img, img_path=img_path, threshold=17000)


outpaint_test_params = [
    ("A132", {"_all": 132}),
    ("A132,U50", {"_all": 132, "up": 50}),
    ("A132,U50,D50", {"_all": 132, "up": 50, "down": 50}),
    ("a132,u50,d50", {"_all": 132, "up": 50, "down": 50}),
    ("all50,up20,down600", {"_all": 50, "up": 20, "down": 600}),
]


@pytest.mark.parametrize(("arg_str", "expected_kwargs"), outpaint_test_params)
def test_outpaint_parse_kwargs(arg_str, expected_kwargs):
    assert outpaint_arg_str_parse(arg_str) == expected_kwargs



import subprocess
import sys
from functools import lru_cache

from setuptools import find_packages, setup

is_for_windows = len(sys.argv) >= 3 and sys.argv[2].startswith("--plat-name=win")

if is_for_windows:
    scripts = None
    entry_points: dict | None = {
        "console_scripts": [
            "imagine=imaginairy.cli.main:imagine_cmd",
            "aimg=imaginairy.cli.main:aimg",
        ],
    }
else:
    scripts = ["imaginairy/cli/bin/aimg", "imaginairy/cli/bin/imagine"]
    entry_points = None


@lru_cache
def get_git_revision_hash() -> str:
    try:
        return (
            subprocess.check_output(["git", "rev-parse", "HEAD"])
            .decode("ascii")
            .strip()
        )
    except FileNotFoundError:
        return "no-git"


revision_hash = get_git_revision_hash()

with open("README.md", encoding="utf-8") as f:
    readme = f.read()
    readme = readme.replace(
        '<img src="',
        f'<img src="https://raw.githubusercontent.com/brycedrennan/imaginAIry/{revision_hash}/',
    )

setup(
    name="imaginAIry",
    author="Bryce Drennan",
    # author_email="b r y p y d o t io",
    version="14.2.0",
    description="AI imagined images. Pythonic generation of images.",
    long_description=readme,
    long_description_content_type="text/markdown",
    project_urls={
        "Documentation": "https://github.com/brycedrennan/imaginAIry/blob/master/README.md",
        "Source": "https://github.com/brycedrennan/imaginAIry",
    },
    packages=find_packages(include=("imaginairy", "imaginairy.*")),
    scripts=scripts,
    entry_points=entry_points,
    package_data={
        "imaginairy": [
            "configs/*.yaml",
            "weight_management/weight_maps/*.json",
            "data/*.*",
            "cli/bin/*.*",
            "http_app/stablestudio/dist/*.*",
            "http_app/stablestudio/dist/assets/*.*",
            "http_app/stablestudio/dist/LICENSE",
            "enhancers/phraselists/*.txt",
            "vendored/clip/*.txt.gz",
            "vendored/clipseg/*.pth",
            "vendored/blip/configs/*.*",
            "vendored/noodle_soup_prompts/*.*",
            "vendored/noodle_soup_prompts/LICENSE",
            "vendored/refiners/foundationals/clip/bpe_simple_vocab_16e6.txt.gz",
        ]
    },
    install_requires=[
        "click>=8.0.0",
        "click-help-colors>=0.9.1",
        "click-shell>=2.0",
        "protobuf != 3.20.2, != 3.19.5",
        "fastapi>=0.70.0",
        "ftfy>=6.0.1",  # for vendored clip
        "torch>=2.1.0",
        # https://numpy.org/neps/nep-0029-deprecation_policy.html
        "numpy>=1.22.0",
        "tqdm>=4.64.0",
        "diffusers>=0.3.0",
        "Pillow>=9.1.0",
        "psutil>5.7.3",
        "omegaconf>=2.1.1",
        "open-clip-torch>=2.0.0",
        "opencv-python>=4.4.0.46",
        # need to migration to 2.0
        "pydantic>=2.3.0",
        # pyparsing used for masking logic and creating text images
        "pyparsing>=3.0.0",
        "requests>=2.28.1",
        # "refiners>=0.2.0",
        "jaxtyping>=0.2.23",  # refiners dependency
        "einops>=0.3.0",
        "safetensors>=0.4.0",
        "scipy>=1.8",
        "termcolor",
        "timm>=0.4.12,!=0.9.0,!=0.9.1",  # for vendored blip
        "torchdiffeq>=0.2.0",
        "torchvision>=0.13.1",
        "transformers>=4.19.2",
        "triton>=2.0.0; sys_platform!='darwin' and platform_machine!='aarch64' and sys_platform == 'linux'",
        "kornia>=0.6",
        "uvicorn>=0.16.0",
        # "xformers>=0.0.22; sys_platform!='darwin' and platform_machine!='aarch64'",
    ],
    # don't specify maximum python versions as it can cause very long dependency resolution issues as the resolver
    # goes back to older versions of packages that didn't specify a maximum
    # https://discuss.python.org/t/requires-python-upper-limits/12663/75
    # https://github.com/brycedrennan/imaginAIry/pull/341#issuecomment-1574723908
    python_requires=">=3.10",
)



import re
from dataclasses import asdict, dataclass, field
from typing import Dict

import torch
from safetensors import safe_open
from torch import device as Device

from imaginairy.utils.log_utils import getLogger

logger = getLogger(__name__)


TensorDict = Dict[str, torch.Tensor]


@dataclass
class WeightTranslationMap:
    name_map: dict[str, str | None] = field(default_factory=dict)
    regex_map: dict[str, str | None] = field(default_factory=dict)
    ignore_prefixes: list[str] = field(default_factory=list)
    source_aliases: dict[str, str] = field(default_factory=dict)
    reshapes: dict[str, tuple[int, ...]] = field(default_factory=dict)

    def load_untranslated_weights(
        self, source_path: str, device: Device | str = "cpu"
    ) -> TensorDict:
        extension = source_path.split(".")[-1]
        if extension in ["pth", "pt", "bin"]:
            source_weights = torch.load(source_path, map_location="cpu")

        elif extension in ["safetensors"]:
            with safe_open(source_path, framework="pt", device=device) as f:  # type: ignore
                source_weights = {k: f.get_tensor(k) for k in f.keys()}  # noqa
        else:
            msg = f"Unsupported extension {extension}"
            raise ValueError(msg)
        return source_weights

    def load_and_translate_weights(
        self, source_path: str, device: Device | str = "cpu"
    ) -> TensorDict:
        source_weights = self.load_untranslated_weights(source_path, device=device)
        return self.translate_weights(source_weights)

    def translate_weights(self, source_weights: TensorDict) -> TensorDict:
        return translate_weights(source_weights=source_weights, weight_map=self)

    def add_regex_replacer(self, find_pattern: str, replace_pattern: str):
        orig_size = len(self.name_map)
        # see which keys of name_map
        find_pattern_c = re.compile(find_pattern)
        matched_keys = [k for k in self.name_map if find_pattern_c.match(k)]
        # print(f"Pattern {find_pattern} matched {len(matched_keys)} keys")
        for k in matched_keys:
            match = find_pattern_c.match(k)
            if not match:
                continue
            data = match.groupdict()
            new_k = render_fstring(replace_pattern, data)
            expected_k = self.name_map[k]
            assert new_k == expected_k
            # print(f"Replacing {k} with {new_k}")
            del self.name_map[k]

        self.regex_map[find_pattern] = replace_pattern
        print(
            f"Adding pattern reduced name_map from {orig_size} to {len(self.name_map)}"
        )

    def save(self, path):
        import json
        import os

        os.makedirs(os.path.dirname(path), exist_ok=True)

        with open(path, "w") as f:
            json.dump(asdict(self), f, indent=4)

    @classmethod
    def load(cls, path):
        import json

        with open(path) as f:
            d = json.load(f)
        return cls(**d)


def check_nan_path(path: str, device):
    from safetensors import safe_open

    with safe_open(path, framework="pt", device=device) as f:  # type: ignore
        for k in f.keys():  # noqa
            if torch.any(torch.isnan(f.get_tensor(k))):
                print(f"Found nan values in {k} of {path}")


def translate_weights(
    source_weights: TensorDict, weight_map: WeightTranslationMap, nan_check=False
) -> TensorDict:
    new_state_dict: TensorDict = {}
    # check source weights for nan
    if nan_check:
        with logger.timed_debug("Checking for nans"):
            for k, v in source_weights.items():
                nan_count = torch.sum(torch.isnan(v)).item()
                if nan_count:
                    msg = (
                        f"Found {nan_count} nan values in {k} of source state dict."
                        " This could indicate the source weights are corrupted and "
                        "need to be re-downloaded. "
                    )
                    logger.warning(msg)

    # print(f"Translating {len(source_weights)} weights")
    # print(f"Using {len(weight_map.name_map)} name mappings")
    # print(source_weights.keys())

    source_weights = flatten_dict(source_weights)

    with logger.timed_debug("Translating exact match keys", hide_below_ms=50):
        for source_key in list(source_weights.keys()):
            source_key = weight_map.source_aliases.get(source_key, source_key)
            try:
                target_key = weight_map.name_map[source_key]
                # print(f"Found {source_prefix} -> {target_prefix}")
            except KeyError:
                continue
            if target_key is None:
                # mapped to None means we ignore it
                source_weights.pop(source_key)
                continue

            new_state_dict[target_key] = source_weights.pop(source_key)

    with logger.timed_debug("Translating prefix matched keys", hide_below_ms=50):
        for source_key in list(source_weights.keys()):
            try:
                source_prefix, suffix = source_key.rsplit(sep=".", maxsplit=1)
            except ValueError:
                # no dots
                continue
            # print(f"Checking {source_prefix} {suffix}")

            source_prefix = weight_map.source_aliases.get(source_prefix, source_prefix)
            try:
                target_prefix = weight_map.name_map[source_prefix]
                # print(f"Found {source_prefix} -> {target_prefix}")
            except KeyError:
                continue
            if target_prefix is None:
                # mapped to None means we ignore it
                source_weights.pop(source_key)
                continue
            else:
                target_key = ".".join([target_prefix, suffix])
                # print(f"Adding {target_key}")
                new_state_dict[target_key] = source_weights.pop(source_key)

    with logger.timed_debug("Translating regex matched keys", hide_below_ms=50):
        for source_key in list(source_weights.keys()):
            try:
                source_prefix, suffix = source_key.rsplit(sep=".", maxsplit=1)
            except ValueError:
                # no dots
                continue
            for pattern, replace_pattern in weight_map.regex_map.items():
                match = re.match(pattern, source_prefix)
                if match:
                    match_data = match.groupdict()
                    new_k = render_fstring(replace_pattern, match_data)
                    new_k = ".".join([new_k, suffix])
                    new_state_dict[new_k] = source_weights.pop(source_key)

    if source_weights:
        msg = f"Unmapped keys: {list(source_weights.keys())}"
        logger.info(msg)
        for k in source_weights:
            if isinstance(source_weights[k], torch.Tensor):
                logger.info(f"  Unmapped key '{k}': {source_weights[k].shape}")
            else:
                logger.info(f"  Unmapped key '{k}': {repr(source_weights[k])[:100]}")

    with logger.timed_debug("reshaping weights", hide_below_ms=50):
        if weight_map.reshapes:
            for key, new_shape in weight_map.reshapes.items():
                if key in new_state_dict:
                    new_state_dict[key] = new_state_dict[key].reshape(new_shape)

    # check for nan values
    if nan_check:
        with logger.timed_debug("Checking for nans", hide_below_ms=50):
            for k in list(new_state_dict.keys()):
                v = new_state_dict[k]
                nan_count = torch.sum(torch.isnan(v)).item()
                if nan_count:
                    logger.warning(
                        f"Found {nan_count} nan values in {k} of converted state dict."
                    )

    return new_state_dict


def flatten_dict(d, sep="."):
    """
    Flattens a nested dictionary into a dictionary with dot-separated keys.
    This function removes items from the original dictionary as they are added to the new one.
    The function uses an iterative approach instead of recursion.

    Parameters:
    d (dict): The dictionary to flatten.
    sep (str): The separator to use between keys.

    Returns:
    dict: A flattened dictionary.
    """
    flat_dict = {}
    stack = [("", d)]

    while stack:
        parent_key, current_dict = stack.pop()
        keys = list(current_dict.keys())  # Create a list of keys to avoid RuntimeError
        for k in keys:
            new_key = f"{parent_key}{sep}{k}" if parent_key else k
            if isinstance(current_dict[k], dict):
                stack.append((new_key, current_dict[k]))
            else:
                flat_dict[new_key] = current_dict.pop(k)

    return flat_dict


def render_fstring(fstring, variables):
    """
    Render a string formatted like an f-string using the provided variables.

    DANGER: This is a security risk if the fstring is user-provided.

    Args:
    fstring (str): The template string with placeholders for variables.
    variables (dict): A dictionary containing the variables to be used in the f-string.

    Returns:
    str: The rendered string with variables substituted.
    """
    # Use locals().update to add the variables to the local scope
    locals().update(variables)

    # Evaluate the f-string using eval with an f-string formatted string
    return eval(f'f"""{fstring}"""')
